"""WISDEM Monopile API."""

__author__ = ["Jake Nunemaker"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from warnings import warn

import openmdao.api as om

from ORBIT import ProjectManager


class Orbit(om.Group):
    """Orbit class for WISDEM API."""

    def initialize(self):
        """Initializes the API connections."""
        self.options.declare("floating", default=False)
        self.options.declare("jacket", default=False)
        self.options.declare("jacket_legs", default=0)

    def setup(self):
        """Define all input variables from all models."""
        self.set_input_defaults("wtiv", "example_wtiv")
        self.set_input_defaults("feeder", "example_feeder")
        #self.set_input_defaults("num_feeders", 1)
        #self.set_input_defaults("num_towing", 1)
        #self.set_input_defaults("num_station_keeping", 3)
        #self.set_input_defaults(
        #    "oss_install_vessel", "example_heavy_lift_vessel",
        #)
        self.set_input_defaults("site_distance", 40.0, units="km")
        self.set_input_defaults("site_distance_to_landfall", 40.0, units="km")
        self.set_input_defaults("interconnection_distance", 40.0, units="km")
        self.set_input_defaults("plant_turbine_spacing", 7)
        self.set_input_defaults("plant_row_spacing", 7)
        self.set_input_defaults("plant_substation_distance", 1, units="km")
        #self.set_input_defaults("num_port_cranes", 1)
        #self.set_input_defaults("num_assembly_lines", 1)
        self.set_input_defaults("takt_time", 170.0, units="h")
        self.set_input_defaults("port_cost_per_month", 2e6, units="USD/mo")
        self.set_input_defaults("construction_insurance", 44.0, units="USD/kW")
        self.set_input_defaults("construction_financing", 183.0, units="USD/kW")
        self.set_input_defaults("contingency", 316.0, units="USD/kW")
        self.set_input_defaults("commissioning_cost_kW", 44.0, units="USD/kW")
        self.set_input_defaults("decommissioning_cost_kW", 58.0, units="USD/kW")
        self.set_input_defaults("site_auction_price", 100e6, units="USD")
        self.set_input_defaults("site_assessment_cost", 50e6, units="USD")
        self.set_input_defaults("construction_plan_cost", 1e6, units="USD")
        self.set_input_defaults("installation_plan_cost", 2.5e5, units="USD")
        self.set_input_defaults("boem_review_cost", 0.0, units="USD")

        self.add_subsystem(
            "orbit",
            OrbitWisdem(
                floating=self.options["floating"],
                jacket=self.options["jacket"],
                jacket_legs=self.options["jacket_legs"],
            ),
            promotes=["*"],
        )


class OrbitWisdem(om.ExplicitComponent):
    """ORBIT-WISDEM Fixed Substructure API."""

    def initialize(self):
        """Initialize the API."""
        self.options.declare("floating", default=False)
        self.options.declare("jacket", default=False)
        self.options.declare("jacket_legs", default=0)

    def setup(self):
        """Define all the inputs."""
        # Inputs
        # self.add_discrete_input(
        #     'weather_file',
        #     'block_island',
        #     desc='Weather file to use for installation times.'
        # )

        # Vessels
        self.add_discrete_input(
            "wtiv",
            "example_wtiv",
            desc=(
                "Vessel configuration to use for installation of foundations"
                " and turbines."
            ),
        )
        self.add_discrete_input(
            "feeder",
            "future_feeder",
            desc="Vessel configuration to use for (optional) feeder barges.",
        )
        self.add_discrete_input(
            "num_feeders",
            1,
            desc=(
                "Number of feeder barges to use for installation of"
                " foundations and turbines."
            ),
        )
        self.add_discrete_input(
            "num_towing",
            1,
            desc=(
                "Number of towing vessels to use for floating platforms that"
                " are assembled at port (with or without the turbine)."
            ),
        )
        self.add_discrete_input(
            "num_station_keeping",
            3,
            desc=(
                "Number of station keeping or AHTS vessels that attach to floating"
                " platforms under tow-out."
            ),
        )
        self.add_discrete_input(
            "oss_install_vessel",
            "example_heavy_lift_vessel",
            desc="Vessel configuration to use for installation of offshore substations.",  # noqa: E501
        )

        # Site
        self.add_input("site_depth", 40.0, units="m", desc="Site depth.")
        self.add_input(
            "site_distance",
            40.0,
            units="km",
            desc="Distance from site to installation port.",
        )
        self.add_input(
            "site_distance_to_landfall",
            50.0,
            units="km",
            desc="Distance from site to landfall for export cable.",
        )
        self.add_input(
            "interconnection_distance",
            3.0,
            units="km",
            desc="Distance from landfall to interconnection.",
        )
        self.add_input(
            "site_mean_windspeed",
            9.0,
            units="m/s",
            desc="Mean windspeed of the site.",
        )

        # Plant
        self.add_discrete_input(
            "number_of_turbines", 60, desc="Number of turbines.",
        )
        self.add_input(
            "plant_turbine_spacing",
            7,
            desc="Turbine spacing in rotor diameters.",
        )
        self.add_input(
            "plant_row_spacing",
            7,
            desc="Row spacing in rotor diameters. Not used in ring layouts.",
        )
        self.add_input(
            "plant_substation_distance",
            1,
            units="km",
            desc="Distance from first turbine in string to substation.",
        )

        # Turbine
        self.add_input(
            "turbine_rating",
            8.0,
            units="MW",
            desc="Rated capacity of a turbine.",
        )
        self.add_input(
            "turbine_rated_windspeed",
            11.0,
            units="m/s",
            desc="Rated windspeed of the turbine.",
        )
        self.add_input(
            "turbine_capex", 1100, units="USD/kW", desc="Turbine CAPEX",
        )
        self.add_input(
            "hub_height", 100.0, units="m", desc="Turbine hub height.",
        )
        self.add_input(
            "turbine_rotor_diameter",
            130,
            units="m",
            desc="Turbine rotor diameter.",
        )
        self.add_input(
            "tower_mass", 400.0, units="t", desc="mass of the total tower.",
        )
        self.add_input(
            "tower_length",
            100.0,
            units="m",
            desc="Total length of the tower.",
        )
        self.add_input(
            "tower_deck_space",
            25.0,
            units="m**2",
            desc=(
                "Deck space required to transport the tower. Defaults to 0 in"
                " order to not be a constraint on installation."
            ),
        )
        self.add_input(
            "nacelle_mass",
            500.0,
            units="t",
            desc="mass of the rotor nacelle assembly (RNA).",
        )
        self.add_input(
            "nacelle_deck_space",
            25.0,
            units="m**2",
            desc=(
                "Deck space required to transport the rotor nacelle assembly"
                " (RNA). Defaults to 0 in order to not be a constraint on"
                " installation."
            ),
        )
        self.add_discrete_input(
            "number_of_blades", 3, desc="Number of blades per turbine.",
        )
        self.add_input(
            "blade_mass", 50.0, units="t", desc="mass of an individual blade.",
        )
        self.add_input(
            "blade_deck_space",
            100.0,
            units="m**2",
            desc=(
                "Deck space required to transport a blade. Defaults to 0 in"
                " order to not be a constraint on installation."
            ),
        )

        # Mooring
        self.add_discrete_input(
            "num_mooring_lines",
            3,
            desc="Number of mooring lines per platform.",
        )
        self.add_input(
            "mooring_line_mass",
            1e4,
            units="kg",
            desc="Total mass of a mooring line",
        )
        self.add_input(
            "mooring_line_diameter",
            0.1,
            units="m",
            desc="Cross-sectional diameter of a mooring line",
        )
        self.add_input(
            "mooring_line_length",
            1e3,
            units="m",
            desc="Unstretched mooring line length",
        )
        self.add_input(
            "anchor_mass", 1e4, units="kg", desc="Total mass of an anchor",
        )
        self.add_input(
            "mooring_line_cost",
            0.5e6,
            units="USD",
            desc="Mooring line unit cost.",
        )
        self.add_input(
            "mooring_anchor_cost",
            0.1e6,
            units="USD",
            desc="Mooring line unit cost.",
        )
        self.add_discrete_input(
            "anchor_type",
            "drag_embedment",
            desc="Number of mooring lines per platform.",
        )

        # Port
        self.add_input(
            "port_cost_per_month",
            2e6,
            units="USD/mo",
            desc="Monthly port costs.",
        )
        self.add_input(
            "takt_time",
            170.0,
            units="h",
            desc="Substructure assembly cycle time when doing assembly at the port.",  # noqa: E501
        )
        self.add_discrete_input(
            "num_assembly_lines",
            1,
            desc="Number of assembly lines used when assembly occurs at the port.",  # noqa: E501
        )
        self.add_discrete_input(
            "num_port_cranes",
            1,
            desc=(
                "Number of cranes used at the port to load feeders / WTIVS"
                " when assembly occurs on-site or assembly cranes when"
                " assembling at port."
            ),
        )

        # Floating Substructures
        self.add_input(
            "floating_substructure_cost",
            10e6,
            units="USD",
            desc="Floating substructure unit cost.",
        )

        # Monopile
        self.add_input(
            "monopile_length",
            100.0,
            units="m",
            desc="Length of monopile (including pile).",
        )
        self.add_input(
            "monopile_diameter", 7.0, units="m", desc="Diameter of monopile.",
        )
        self.add_input(
            "monopile_mass",
            900.0,
            units="t",
            desc="mass of an individual monopile.",
        )
        self.add_input(
            "monopile_cost", 4e6, units="USD", desc="Monopile unit cost.",
        )

        # Jacket
        self.add_input(
            "jacket_length",
            65.0,
            units="m",
            desc="Length/height of jacket (including pile/buckets).",
        )
        self.add_input(
            "jacket_mass",
            900.0,
            units="t",
            desc="mass of an individual jacket.",
        )
        self.add_input(
            "jacket_cost", 4e6, units="USD", desc="Jacket unit cost.",
        )
        self.add_input(
            "jacket_r_foot",
            10.0,
            units="m",
            desc="Radius of jacket legs at base from centeroid.",
        )

        # Generic fixed-bottom
        self.add_input(
            "transition_piece_mass",
            250.0,
            units="t",
            desc="mass of an individual transition piece.",
        )
        self.add_input(
            "transition_piece_deck_space",
            25.0,
            units="m**2",
            desc=(
                "Deck space required to transport a transition piece."
                " Defaults to 0 in order to not be a constraint on"
                " installation."
            ),
        )
        self.add_input(
            "transition_piece_cost",
            1.5e6,
            units="USD",
            desc="Transition piece unit cost.",
        )

        # Project
        self.add_input(
            "construction_insurance",
            44.0,
            units="USD/kW",
            desc="Cost for construction insurance",
        )
        self.add_input(
            "construction_financing",
            183.0,
            units="USD/kW",
            desc="Cost for construction financing",
        )
        self.add_input(
            "contingency",
            316.0,
            units="USD/kW",
            desc="Cost in case of contingency",
        )
        self.add_input(
            "site_auction_price",
            100e6,
            units="USD",
            desc="Cost to secure site lease",
        )
        self.add_input(
            "site_assessment_cost",
            50e6,
            units="USD",
            desc="Cost to execute site assessment",
        )
        self.add_input(
            "construction_plan_cost",
            1e6,
            units="USD",
            desc="Cost to do construction planning",
        )
        self.add_input(
            "installation_plan_cost",
            2.5e5,
            units="USD",
            desc="Cost to do construction planning",
        )
        self.add_input(
            "boem_review_cost",
            0.0,
            units="USD",
            desc=(
                "Cost for additional review by U.S. Dept of Interior Bureau"
                " of Ocean Energy Management (BOEM)"
            ),
        )
        self.add_input("commissioning_cost_kW", 44.0, units="USD/kW", desc="Commissioning cost.")
        self.add_input("decommissioning_cost_kW", 58.0, units="USD/kW", desc="Decommissioning cost.")

        # Outputs
        # Totals
        self.add_output(
            "bos_capex",
            0.0,
            units="USD",
            desc="Sum of system and installation capex",
        )
        self.add_output(
            "soft_capex",
            0.0,
            units="USD",
            desc="Project costs associated with commissioning, decommissioning and financing",
        )
        self.add_output(
            "project_capex",
            0.0,
            units="USD",
            desc="costs associated with the lease area, "+
            "the development of the construction operations plan,"+
            "and any environmental review and other upfront project costs."
        )
        self.add_output(
            "total_capex",
            0.0,
            units="USD",
            desc="Total capex of bos + soft + project",
        )
        self.add_output(
            "total_capex_kW",
            0.0,
            units="USD/kW",
            desc="Total capex of bos + soft + project per rated project capacity in kW",
        )
        self.add_output(
            "installation_time",
            0.0,
            units="h",
            desc="Total balance of system installation time.",
        )
        self.add_output(
            "installation_capex",
            0.0,
            units="USD",
            desc="Total balance of system installation cost.",
        )

    def compile_orbit_config_file(
        self, inputs, outputs, discrete_inputs, discrete_outputs,
    ):
        """Compiles the ORBIT configuration dictionary."""

        floating_flag = self.options["floating"]
        jacket_flag = self.options["jacket"]

        config = {
            # Vessels
            "wtiv": (
                "floating_heavy_lift_vessel"
                if floating_flag
                else discrete_inputs["wtiv"]
            ),
            "array_cable_install_vessel": "example_cable_lay_vessel",
            "array_cable_bury_vessel": "example_cable_lay_vessel",
            "export_cable_install_vessel": "example_cable_lay_vessel",
            "export_cable_bury_vessel": "example_cable_lay_vessel",
            # Site/plant
            "site": {
                "depth": float(inputs["site_depth"][0]),
                "distance": float(inputs["site_distance"][0]),
                "distance_to_landfall": float(
                    inputs["site_distance_to_landfall"][0]
                ),
                "mean_windspeed": float(inputs["site_mean_windspeed"][0]),
            },
            "plant": {
                "layout": "grid",
                "num_turbines": int(discrete_inputs["number_of_turbines"]),
                "row_spacing": float(inputs["plant_row_spacing"][0]),
                "turbine_spacing": float(inputs["plant_turbine_spacing"][0]),
                "substation_distance": float(
                    inputs["plant_substation_distance"][0]
                ),
            },
            # Turbine + components
            "turbine": {
                "hub_height": float(inputs["hub_height"][0]),
                "rotor_diameter": float(inputs["turbine_rotor_diameter"][0]),
                "turbine_rating": float(inputs["turbine_rating"][0]),
                "rated_windspeed": float(inputs["turbine_rated_windspeed"][0]),
                "tower": {
                    "type": "Tower",
                    "deck_space": float(inputs["tower_deck_space"][0]),
                    "mass": float(inputs["tower_mass"][0]),
                    "length": float(inputs["tower_length"][0]),
                },
                "nacelle": {
                    "type": "Nacelle",
                    "deck_space": float(inputs["nacelle_deck_space"][0]),
                    "mass": float(inputs["nacelle_mass"][0]),
                },
                "blade": {
                    "type": "Blade",
                    "number": int(discrete_inputs["number_of_blades"]),
                    "deck_space": float(inputs["blade_deck_space"][0]),
                    "mass": float(inputs["blade_mass"][0]),
                },
            },
            # Substructure components
            "transition_piece": {
                "type": "Transition Piece",
                "deck_space": float(inputs["transition_piece_deck_space"][0]),
                "mass": float(inputs["transition_piece_mass"][0]),
                "unit_cost": float(inputs["transition_piece_cost"][0]),
            },
            # Electrical
            "array_system_design": {
                "cables": ["XLPE_630mm_66kV", "XLPE_185mm_66kV"],
            },
            "export_system_design": {
                "cables": "XLPE_1000mm_220kV",
                "percent_added_length": 0.1,
                "landfall": {
                    "interconnection_distance": float(
                        inputs["interconnection_distance"][0],
                    ),
                },
            },
            # Phase Specific
            "OffshoreSubstationInstallation": {
                "oss_install_vessel": (
                    "floating_heavy_lift_vessel"
                    if floating_flag
                    else "example_heavy_lift_vessel"
                ),
                "feeder": (
                    "floating_barge" if floating_flag else "future_feeder"
                ),
                "num_feeders": int(discrete_inputs["num_feeders"]),
            },
            # Project development costs
            "project_parameters": {
                "construction_insurance": float(inputs["construction_insurance"][0]),
                "construction_financing": float(inputs["construction_financing"][0]),
                "installation_contingency": 0.5*float(inputs["contingency"][0]),
                "procurement_contingency": 0.5*float(inputs["contingency"][0]),
                "site_auction_price": float(inputs["site_auction_price"][0]),
                "site_assessment_cost": float(inputs["site_assessment_cost"][0]),
                "construction_plan_cost": float(inputs["construction_plan_cost"][0]),
                "installation_plan_cost": float(inputs["installation_plan_cost"][0]),
                "boem_review_cost": float(inputs["boem_review_cost"][0]),
                "commissioning": float(inputs["commissioning_cost_kW"][0]),
                "decommissioning": float(inputs["decommissioning_cost_kW"][0]),
                "turbine_capex": float(inputs["turbine_capex"][0]),
            },
            # Phases
            # Putting monopile or semisub here would override the inputs
            # we assume to get from WISDEM
            "design_phases": [
                # 'MonopileDesign',
                # 'SemiSubmersibleDesign',
                # 'MooringSystemDesign',
                # 'ScourProtectionDesign',
                "ArraySystemDesign",
                "ExportSystemDesign",
                "OffshoreSubstationDesign",
            ],
        }

        if "landfall" in config and "interconnection_distance" in config["landfall"]:
            warn(
                "landfall dictionary will be deprecated and moved"
                " into [export_system_design][landfall].",
                DeprecationWarning,
                stacklevel=2,
            )

        if "export_system_design" in config and "interconnection_distance" in config["export_system_design"]:
            warn(
                "[export_system][interconnection_distance] will be deprecated"
                " and moved to"
                " [export_system_design][landfall][interconnection_distance].",
                DeprecationWarning,
                stacklevel=2,
            )

        # Unique design phases
        if floating_flag:
            config["install_phases"] = {
                "ExportCableInstallation": 0,
                "OffshoreSubstationInstallation": 0,
                "MooringSystemInstallation": 0,
                "MooredSubInstallation": ("MooringSystemInstallation", 0.25),
                "ArrayCableInstallation": ("MooredSubInstallation", 0.25),
            }
        else:
            fixedStr = (
                "JacketInstallation" if jacket_flag else "MonopileInstallation"
            )

            if jacket_flag:
                monopile = config.get("monopile", {})
                monopile["diameter"] = 10
                config["monopile"] = monopile

            config["design_phases"] += ["ScourProtectionDesign"]
            config["install_phases"] = {
                "ExportCableInstallation": 0,
                "OffshoreSubstationInstallation": 0,
                "ArrayCableInstallation": 0,
                fixedStr: 0,
                "ScourProtectionInstallation": 0,
                "TurbineInstallation": (fixedStr, 0.25),
            }

        # Unique vessels
        if floating_flag:
            vessels = {
                "ahts_vessel": "example_ahts_vessel",
                "towing_vessel": "example_towing_vessel",
                "mooring_install_vessel": "example_support_vessel",
                "towing_vessel_groups": {
                    "towing_vessels": int(discrete_inputs["num_towing"]),
                    "ahts_vessels": int(
                        discrete_inputs["num_station_keeping"]
                    ),
                },
            }
        else:
            vessels = {
                "feeder": discrete_inputs["feeder"],
                "num_feeders": int(discrete_inputs["num_feeders"]),
                "spi_vessel": "example_scour_protection_vessel",
            }
        config.update(vessels)

        # Unique support structure design/assembly
        if floating_flag:
            config["port"] = {
                "sub_assembly_lines": int(
                    discrete_inputs["num_assembly_lines"]
                ),
                "turbine_assembly_cranes": int(
                    discrete_inputs["num_port_cranes"]
                ),
                "monthly_rate": float(inputs["port_cost_per_month"][0]),
            }

            config["substructure"] = {
                "takt_time": float(inputs["takt_time"][0]),
                "unit_cost": float(inputs["floating_substructure_cost"][0]),
                "towing_speed": 6.0,  # km/h
            }

            anchorstr_in = discrete_inputs["anchor_type"].lower()
            if anchorstr_in.find("drag") >= 0:
                anchorstr = "Drag Embedment"
            elif anchorstr_in.find("suction") >= 0:
                anchorstr = "Suction Pile"

            config["mooring_system"] = {
                "num_lines": int(discrete_inputs["num_mooring_lines"]),
                "line_mass": 1e-3 * float(inputs["mooring_line_mass"][0]),
                "line_diam": float(inputs["mooring_line_diameter"][0]),
                "line_length": float(inputs["mooring_line_length"][0]),
                "line_cost": float(inputs["mooring_line_cost"][0]),
                "anchor_mass": 1e-3 * float(inputs["anchor_mass"][0]),
                "anchor_type": anchorstr,
                "anchor_cost": float(inputs["mooring_anchor_cost"][0]),
            }
        else:
            config["port"] = {
                "num_cranes": int(discrete_inputs["num_port_cranes"]),
                "monthly_rate": float(inputs["port_cost_per_month"][0]),
            }

            config["scour_protection_design"] = {
                "cost_per_tonne": 20,
            }

            if jacket_flag:
                config["jacket"] = {
                    "type": "Jacket",
                    "height": float(inputs["jacket_length"][0]),
                    "num_legs": int(self.options["jacket_legs"]),
                    "deck_space": 4 * float(inputs["jacket_r_foot"][0]) ** 2,
                    "mass": float(inputs["jacket_mass"][0]),
                    "unit_cost": float(inputs["jacket_cost"][0]),
                }
            else:
                config["monopile"] = {
                    "type": "Monopile",
                    "length": float(inputs["monopile_length"][0]),
                    "diameter": float(inputs["monopile_diameter"][0]),
                    "deck_space": 0.25
                    * float(
                        inputs["monopile_diameter"][0]
                        * inputs["monopile_length"][0]
                    ),
                    "mass": float(inputs["monopile_mass"][0]),
                    "unit_cost": float(inputs["monopile_cost"][0]),
                }

        self._orbit_config = config
        return config

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        """Creates and runs the project, then gathers the results."""

        config = self.compile_orbit_config_file(
            inputs, outputs, discrete_inputs, discrete_outputs,
        )

        project = ProjectManager(config)
        project.run()

        # The ORBIT version of total_capex includes turbine capex, so we do our own sum of
        # the parts here that wisdem doesn't account for
        capacity_kW = 1e3 * inputs["turbine_rating"] * discrete_inputs["number_of_turbines"]
        outputs["bos_capex"] = project.bos_capex
        outputs["soft_capex"] = project.soft_capex
        outputs["project_capex"] = project.project_capex
        outputs["total_capex"] = project.bos_capex + project.soft_capex + project.project_capex
        outputs["total_capex_kW"] = outputs["total_capex"] / capacity_kW
        outputs["installation_time"] = project.installation_time
        outputs["installation_capex"] = project.installation_capex



import numpy as np
import math
from wisdem.landbosse.model.WeatherDelay import WeatherDelay as WD
import traceback
from wisdem.landbosse.model.CostModule import CostModule
import pandas as pd


class SitePreparationCost(CostModule):
    """
    **SitePreparationCost.py**

    - Created by Annika Eberle and Owen Roberts on Apr. 3, 2018

    - Refactored by Parangat Bhaskar and Alicia Key on Jun 3, 2019



    **Calculates cost of constructing roads for land-based wind projects:**

    - Get terrain complexity

    - Get turbine spacing

    - Get road width

    - Get number of turbines

    - Get turbine rating

    - Get duration of construction*  #todo: add to process diagram

    - Get road length

    - Get weather data

    - Get road thickness

    - Calculate volume of road based on road thickness, road width, and road length

    - Calculate road labor and equipment costs by operation and type using RSMeans data

    - Calculate man hours and equipment hours for compaction of soil based on road length, road thickness, soil type, road width, and equipment size

    - Calculate man hours and equipment hours for mass material movement based on land cover, terrain complexity, and road length

    - Calculate man hours and equipment hours for rock placement based on equipment size, distance to quarry, and volume of road

    - Calculate man hours and equipment hours for compaction of rock based on road length, road thickness, and rock type

    - Calculate man hours and equipment hours for final grading based on road length

    - Calculate quantity of materials based on volume of materials

    - Calculate material costs by type

    - Calculate material costs using quantity of materials by material type and material prices by material type

    - Sum road costs over all operations and material types to get total costs by type of cost (e.g., material vs. equipment)

    - Return total road costs by type of cost


    **Keys in the input dictionary are the following:**

    - road_length

    - road_width

    - road_thickness

    - crane_width

    - num_access_roads  #TODO: Add to excel inputs sheet

    - num_turbines

    - rsmeans (dataframe)

    - duration_construction

    - wind_delays

    - wind_delay_time

    - material_price (dataframe)

    - rsmeans_per_diem

    - rotor_diameter_m

    - turbine_spacing_rotor_diameters


    **Keys in the output dictionary are the following:**

    - road_length_m

    - road_volume_m3

    - depth_to_subgrade_m

    - crane_path_widthm

    - road_thickess_m

    - road_width_m

    - road_width_m

    - material_volume_cubic_yards

    - road_construction_time

    - topsoil_volume

    - embankment_volume_crane

    - embankment_volume_road

    - rough_grading_area

    - material_needs (dataframe)

    - operation_data (dataframe)

    - total_road_cost



    """

    def __init__(self, input_dict, output_dict, project_name):
        """
        Parameters
        ----------
        input_dict : dict
            The input dictionary with key value pairs described in the
            class documentation

        output_dict : dict
            The output dictionary with key value pairs as found on the
            output documentation.
        """

        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name

        # Road quality and fraction of roads that are
        self.fraction_new_roads = self.input_dict["fraction_new_roads"]
        self.road_quality = self.input_dict["road_quality"]

        # Conversion factors. Making this data private (hidden from outside of this class):
        self._meters_per_foot = 0.3
        self._meters_per_inch = 0.025
        self._cubic_yards_per_cubic_meter = 1.30795
        self._square_feet_per_square_meter = 10.7639

        # cubic meters for crane pad and maintenance ring for each turbine
        # (from old BOS model - AI - Access Roads & Site Imp. tab cell J33)
        self._crane_pad_volume = 125

        # conversion factor for converting packed cubic yards to loose cubic yards
        # material volume is about 1.4 times greater than road volume due to compaction
        self._yards_loose_per_yards_packed = 1.39

        # calculated road properties
        self._lift_depth_m = 0.2

    def calculate_road_properties(self, road_properties_input, road_properties_output):
        """
        Calculates the volume of road materials need based on length, width, and thickness of roads

        Parameters
        ----------

        int num_turbines
            number of turbines in wind farm

        unitless turbine_spacing_rotor_diameters
            Immediate spacing between two turbines as a function of rotor diameter

        int rotor_diameter_m
            Rotor diameter

        float crane_width
            Crane width in meters

        float road_thickness
            Road thickness in inches

        float road_width_ft
            Road width in feet



        Returns
        ----------

        road_length_m
            Calculated road length (in m)

        road_volume_m3
            Calculated road volume (in m^3)

        material_volume_cubic_yards
            Material volume required (in cubic yards) based on road volume


        """

        if road_properties_input["road_distributed_wind"] == True or road_properties_input["turbine_rating_MW"] < 0.1:
            road_properties_output["road_volume"] = (
                road_properties_input["road_length_adder_m"]
                * (road_properties_input["road_width_ft"] * self._meters_per_foot)
                * (road_properties_input["road_thickness"] * self._meters_per_inch)
            )  # units of cubic meters
        else:
            road_properties_output["road_length_m"] = (
                (road_properties_input["num_turbines"] - 1)
                * road_properties_input["turbine_spacing_rotor_diameters"]
                * road_properties_input["rotor_diameter_m"]
            ) + road_properties_input["road_length_adder_m"]
            road_properties_output["road_width_m"] = road_properties_input["road_width_ft"] * self._meters_per_foot
            road_properties_output["road_volume"] = (
                road_properties_output["road_length_m"]
                * (road_properties_input["road_width_ft"] * self._meters_per_foot)
                * (road_properties_input["road_thickness"] * self._meters_per_inch)
            )  # units of cubic meters

        road_properties_output["road_volume_m3"] = (
            road_properties_output["road_volume"] + self._crane_pad_volume * road_properties_input["num_turbines"]
        )  # in cubic meters

        road_properties_output["depth_to_subgrade_m"] = 0.1

        road_properties_output["crane_path_width_m"] = road_properties_input["crane_width"] + 1.5

        # todo: replace with actual crane path width from erection module
        road_properties_output["road_thickness_m"] = road_properties_input["road_thickness"] * self._meters_per_inch
        road_properties_output["road_width_m"] = road_properties_input["road_width_ft"] * self._meters_per_foot
        road_properties_output["material_volume_cubic_yards"] = (
            road_properties_output["road_volume_m3"]
            * self._cubic_yards_per_cubic_meter
            * self._yards_loose_per_yards_packed
        )  # todo: output_dict material volume

        return road_properties_output

    def estimate_construction_time(self, estimate_construction_time_input, estimate_construction_time_output):
        """
        Estimates construction time of roads for entire project.

        Parameters
        ----------
        float crane_path_width_m
            Width of crane path (in m)

        float road_length_m
            Road length (in m)

        float depth_to_subgrade_m
            Depth to subgarde (in m)

        float road_volume
            Road volume (in m^3)

        float road_thickness_m
            Road thickness (in m)



        Returns
        ----------

        pd.DataFrame operation_data
            Dataframe which conatains following outputs:

            -  Number of days required for construction

            - Number of crews required to complete roads construction in specified construction time

            - Cost of labor and equipment rental prior to weather delays

        """
        throughput_operations = estimate_construction_time_input["rsmeans"]

        # TODO: Figure out where 'construct_duration' gets read in.
        estimate_construction_time_output["road_construction_time"] = (
            estimate_construction_time_input["construct_duration"] * 1 / 5
        )  # assumes road construction occurs for 1/5 of project time

        # Main switch between small DW wind and (utility scale + distributed wind)
        # select operations for roads module that have data
        if estimate_construction_time_input["turbine_rating_MW"] >= 0.1:
            operation_data = throughput_operations.where(throughput_operations["Module"] == "Roads").dropna(thresh=4)
        else:
            operation_data = throughput_operations.where(throughput_operations["Module"] == "Small DW Roads").dropna(
                thresh=4
            )
            operation_data = operation_data.dropna(subset=["Units"])

        # create list of unique material units for operations
        list_units = operation_data["Units"].unique()

        if (
            estimate_construction_time_input["road_distributed_wind"] == True
            and estimate_construction_time_input["turbine_rating_MW"] >= 0.1
        ):
            estimate_construction_time_output["topsoil_volume"] = (
                estimate_construction_time_input["site_prep_area_m2"]
                * (estimate_construction_time_output["depth_to_subgrade_m"])
                * self._cubic_yards_per_cubic_meter
            )  # units: cubic yards
            estimate_construction_time_output["embankment_volume_crane"] = estimate_construction_time_output[
                "topsoil_volume"
            ]  # units: cubic yards
            estimate_construction_time_output["embankment_volume_road"] = estimate_construction_time_output[
                "topsoil_volume"
            ]  # units: cubic yards
            estimate_construction_time_output["rough_grading_area"] = (
                estimate_construction_time_input["site_prep_area_m2"] * 10.76391
            ) / 100000  # where 10.76391 sq ft = 1 sq m

        elif (
            estimate_construction_time_input["road_distributed_wind"] == True
            and estimate_construction_time_input["turbine_rating_MW"] < 0.1
        ):
            estimate_construction_time_output["topsoil_volume"] = (
                estimate_construction_time_input["road_length_adder_m"]
                * (estimate_construction_time_input["road_width_ft"] * 0.3048)
                * (estimate_construction_time_input["road_thickness"] * 0.0254)
                * 1.308
            )  # Units: CY (where 1 m3 = 1.308 CY)

        else:
            estimate_construction_time_output["topsoil_volume"] = (
                (estimate_construction_time_output["crane_path_width_m"])
                * estimate_construction_time_output["road_length_m"]
                * (estimate_construction_time_output["depth_to_subgrade_m"])
                * self._cubic_yards_per_cubic_meter
            )  # units: cubic yards
            estimate_construction_time_output["embankment_volume_crane"] = (
                (estimate_construction_time_output["crane_path_width_m"])
                * estimate_construction_time_output["road_length_m"]
                * (estimate_construction_time_output["depth_to_subgrade_m"])
                * self._cubic_yards_per_cubic_meter
            )  # units: cubic yards
            estimate_construction_time_output["embankment_volume_road"] = (
                estimate_construction_time_output["road_volume"]
                * self._cubic_yards_per_cubic_meter
                * math.ceil(estimate_construction_time_output["road_thickness_m"] / self._lift_depth_m)
            )  # units: cubic yards road
            estimate_construction_time_output["rough_grading_area"] = (
                estimate_construction_time_output["road_length_m"]
                * estimate_construction_time_output["road_width_m"]
                * self._square_feet_per_square_meter
                * math.ceil(estimate_construction_time_output["road_thickness_m"] / self._lift_depth_m)
                / 100000
            )  # Units: Each (100,000 square feet)

        if estimate_construction_time_input["turbine_rating_MW"] >= 0.1:
            material_quantity_dict = {
                "cubic yard": estimate_construction_time_output["topsoil_volume"],
                "embankment cubic yards crane": estimate_construction_time_output["embankment_volume_crane"],
                "embankment cubic yards road": estimate_construction_time_output["embankment_volume_road"],
                "loose cubic yard": estimate_construction_time_output["material_volume_cubic_yards"],
                "Each (100000 square feet)": estimate_construction_time_output["rough_grading_area"],
            }
        else:  # small DW
            material_quantity_dict = {
                "cubic yard": estimate_construction_time_output["topsoil_volume"],
                "embankment cubic yards crane": estimate_construction_time_output["topsoil_volume"],
                "loose cubic yard": estimate_construction_time_output["topsoil_volume"],
                "embankment cubic yards road": estimate_construction_time_output["topsoil_volume"],
            }

        material_needs = pd.DataFrame(
            data=[[unit, material_quantity_dict[unit]] for unit in list_units],
            columns=["Units", "Quantity of material"],
        )

        estimate_construction_time_output["material_needs"] = material_needs

        # join material needs with operational data to compute costs
        operation_data = pd.merge(operation_data, material_needs, on=["Units"], sort=True).dropna(thresh=3)
        operation_data = operation_data.where((operation_data["Daily output"]).isnull() == False).dropna(thresh=4)

        # calculate operational parameters and estimate costs without weather delays
        operation_data["Number of days"] = operation_data["Quantity of material"] / operation_data["Daily output"]
        operation_data["Number of crews"] = np.ceil(
            (operation_data["Number of days"] / 30.0) / estimate_construction_time_output["road_construction_time"]
        )
        operation_data["Cost USD without weather delays"] = (
            operation_data["Quantity of material"] * operation_data["Rate USD per unit"]
        )

        # if more than one crew needed to complete within construction duration then assume that all construction happens
        # within that window and use that time frame for weather delays; if not, use the number of days calculated
        operation_data["time_construct_bool"] = (
            operation_data["Number of days"] > estimate_construction_time_output["road_construction_time"] * 30.0
        )
        boolean_dictionary = {True: estimate_construction_time_output["road_construction_time"] * 30.0, False: np.nan}
        operation_data["time_construct_bool"] = operation_data["time_construct_bool"].map(boolean_dictionary)
        operation_data["Time construct days"] = operation_data[["time_construct_bool", "Number of days"]].min(axis=1)
        num_days = operation_data["Time construct days"].max()

        # pull out management data
        if estimate_construction_time_input["turbine_rating_MW"] >= 0.1:
            crew_cost = self.input_dict["crew_cost"]
            crew = self.input_dict["crew"][self.input_dict["crew"]["Crew type ID"].str.contains("M0")]
            management_crew = pd.merge(crew_cost, crew, on=["Labor type ID"], sort=True)
            management_crew = management_crew.assign(
                per_diem_total=management_crew["Per diem USD per day"] * management_crew["Number of workers"] * num_days
            )
            management_crew = management_crew.assign(
                hourly_costs_total=management_crew["Hourly rate USD per hour"]
                * self.input_dict["hour_day"][self.input_dict["time_construct"]]
                * num_days
            )
            management_crew = management_crew.assign(
                total_crew_cost_before_wind_delay=management_crew["per_diem_total"]
                + management_crew["hourly_costs_total"]
            )
            self.output_dict["management_crew"] = management_crew

            self.output_dict["managament_crew_cost_before_wind_delay"] = management_crew[
                "total_crew_cost_before_wind_delay"
            ].sum()

        estimate_construction_time_output["operation_data"] = operation_data

        return operation_data

    def calculate_weather_delay(self, weather_delay_input_data, weather_delay_output_data):
        """
        Calculates wind delay for roads.

        """

        # construct WeatherDelay module
        WD(weather_delay_input_data, weather_delay_output_data)

        # compute weather delay
        wind_delay = pd.DataFrame(weather_delay_output_data["wind_delays"])

        # if greater than 4 hour delay, then shut down for full day (10 hours)
        wind_delay[(wind_delay > 4)] = 10
        weather_delay_output_data["wind_delay_time"] = float(wind_delay.sum().iloc[0])

        return weather_delay_output_data

    def new_and_existing_total_road_cost(self, new_road_cost_by_type):
        """
        Calculates total road cost, with contribution from new and existing roads. Refer to eq. 3.3.10 in Technical Report.

        Returns
        -------
        float new_and_existing_total_road_cost_USD_df
            Sum total cost in USD of new and existing roads.
        """

        f_new = (
            self.fraction_new_roads
        )  # fraction of new roads that will be built (default assumes 33% new). TODO: Change to user input.
        r_q = (
            self.road_quality
        )  # is a non-dimensional representation of road quality. Default assumes 0.6, which is representative of average road conditions. TODO: Change to user input.
        cost_existing_roads = (
            0.5 * new_road_cost_by_type
        )  # cost of installing existing roads for the entire project. We currently assume default case to equal 50 % of cost of building new roads.
        new_and_existing_total_road_cost = (f_new * new_road_cost_by_type) + ((1 - f_new) * r_q * cost_existing_roads)
        return new_and_existing_total_road_cost

    def calculate_costs(self, calculate_cost_input_dict, calculate_cost_output_dict):
        """
        Function to calculate total labor, equipment, material, mobilization, and anyother associated costs after
        factoring in weather delays.


        Parameters
        ----------
        pd.Dataframe RSMeans
            Dataframe containing labor and equipment rental costs

        pd.DataFrame operation_data
            DataFrame containing estimates for total roads construction time and cost


        Returns
        ----------

        pd.DataFrame total_road_cost
            Dataframe containing following calculated outputs (after weather delay considerations):

            - Total labor cost

            - Totoal material cost

            - Total equipment rental cost

            - Total mobilization cost

            - Any other related costs


        """
        rsmeans = calculate_cost_input_dict["rsmeans"]

        material_name = rsmeans["Material type ID"].where(rsmeans["Module"] == "Roads").dropna().unique()

        material_vol = pd.DataFrame(
            [[material_name[0], calculate_cost_output_dict["material_volume_cubic_yards"], "Loose cubic yard"]],
            columns=["Material type ID", "Quantity of material", "Units"],
        )

        material_data = pd.merge(
            material_vol, calculate_cost_input_dict["material_price"], on=["Material type ID"], sort=True
        )
        material_data["Cost USD"] = material_data["Quantity of material"] * pd.to_numeric(
            material_data["Material price USD per unit"]
        )

        # New roads material cost:
        material_cost_of_new_roads = material_data["Quantity of material"].iloc[0] * pd.to_numeric(
            material_data["Material price USD per unit"].iloc[0]
        )

        # New + old roads material cost:
        if calculate_cost_input_dict["turbine_rating_MW"] >= 0.1:
            material_cost_of_old_and_new_roads = self.new_and_existing_total_road_cost(material_cost_of_new_roads)
            material_costs = pd.DataFrame(
                [["Materials", float(material_cost_of_old_and_new_roads), "Roads"]],
                columns=["Type of cost", "Cost USD", "Phase of construction"],
            )
        else:
            material_cost_of_old_and_new_roads = material_cost_of_new_roads
            material_costs = pd.DataFrame(
                [["Materials", float(material_cost_of_old_and_new_roads), "Small DW Roads"]],
                columns=["Type of cost", "Cost USD", "Phase of construction"],
            )

        operation_data = self.estimate_construction_time(calculate_cost_input_dict, calculate_cost_output_dict)

        wind_delay_fraction = (
            calculate_cost_output_dict["wind_delay_time"] / calculate_cost_input_dict["operational_hrs_per_day"]
        ) / operation_data["Time construct days"].max(skipna=True)
        # check if wind_delay_fraction is greater than 1, which would mean weather delays are longer than they can possibily be for the input data
        if wind_delay_fraction > 1:
            raise ValueError("{}: Error: Wind delay greater than 100%".format(type(self).__name__))
        calculate_cost_output_dict["wind_multiplier"] = 1.0 / (1.0 - wind_delay_fraction)

        per_diem = (
            operation_data["Number of workers"]
            * operation_data["Number of crews"]
            * (operation_data["Time construct days"] + np.ceil(operation_data["Time construct days"] / 7.0))
            * calculate_cost_input_dict["rsmeans_per_diem"]
        )
        labor_per_diem = per_diem.dropna()

        # calculate_cost_output_dict['Total per diem (USD)'] = per_diem.sum()
        labor_equip_data = pd.merge(
            operation_data[["Operation ID", "Units", "Quantity of material"]],
            rsmeans,
            on=["Units", "Operation ID"],
            sort=True,
        )
        # Calculating labor costs:
        if calculate_cost_input_dict["turbine_rating_MW"] >= 0.1:
            labor_equip_data["Calculated per diem"] = per_diem
            labor_data = labor_equip_data[labor_equip_data["Type of cost"] == "Labor"].copy()
        else:
            labor_equip_data["Calculated per diem"] = 0.0
            # labor_data = labor_equip_data[labor_equip_data['Type of cost'] == 'Labor'].copy()
            labor_data = labor_equip_data[labor_equip_data["Module"] == "Small DW Roads"].copy()

        overtime_multiplier = calculate_cost_input_dict["overtime_multiplier"]
        wind_multiplier = calculate_cost_output_dict["wind_multiplier"]

        labor_data["Cost USD"] = (
            (labor_data["Quantity of material"] * labor_data["Rate USD per unit"] * overtime_multiplier)
            + labor_per_diem
        ) * wind_multiplier

        if calculate_cost_input_dict["road_distributed_wind"] and calculate_cost_input_dict["turbine_rating_MW"] >= 0.1:

            labor_for_new_roads_cost_usd = (
                labor_data["Cost USD"].sum() + calculate_cost_output_dict["managament_crew_cost_before_wind_delay"]
            )

            labor_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(labor_for_new_roads_cost_usd)
            labor_costs = pd.DataFrame(
                [["Labor", float(labor_for_new_and_old_roads_cost_usd), "Roads"]],
                columns=["Type of cost", "Cost USD", "Phase of construction"],
            )

        elif (
            calculate_cost_input_dict["road_distributed_wind"] and calculate_cost_input_dict["turbine_rating_MW"] < 0.1
        ):  # small DW

            labor_for_new_roads_cost_usd = labor_data["Cost USD"].sum()
            labor_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(labor_for_new_roads_cost_usd)

            labor_costs = pd.DataFrame(
                [["Labor", float(labor_for_new_and_old_roads_cost_usd), "Small DW Roads"]],
                columns=["Type of cost", "Cost USD", "Phase of construction"],
            )

        else:
            labor_for_new_roads_cost_usd = (
                labor_data["Cost USD"].sum() + calculate_cost_output_dict["managament_crew_cost_before_wind_delay"]
            )

            labor_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(labor_for_new_roads_cost_usd)

            labor_costs = pd.DataFrame(
                [["Labor", float(labor_for_new_and_old_roads_cost_usd), "Roads"]],
                columns=["Type of cost", "Cost USD", "Phase of construction"],
            )

        # Filter out equipment costs from rsmeans tab:
        if calculate_cost_input_dict["turbine_rating_MW"] >= 0.1:
            equipment_data = labor_equip_data[labor_equip_data["Type of cost"] == "Equipment rental"].copy()
        else:
            equipment_data = labor_equip_data[labor_equip_data["Module"] == "Small DW Roads"].copy()
            equipment_data = equipment_data[equipment_data["Type of cost"] == "Equipment rental"].copy()

        equipment_data["Cost USD"] = (
            equipment_data["Quantity of material"] * equipment_data["Rate USD per unit"]
        ) * calculate_cost_output_dict["wind_multiplier"]

        # if rental cost is < half day minimum:
        if equipment_data["Cost USD"].sum() < 460:
            equip_for_new_roads_cost_usd = 460
        else:
            equip_for_new_roads_cost_usd = equipment_data["Cost USD"].sum()

        if calculate_cost_input_dict["turbine_rating_MW"] >= 0.1:
            equip_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(equip_for_new_roads_cost_usd)
            equipment_costs = pd.DataFrame(
                [["Equipment rental", float(equip_for_new_and_old_roads_cost_usd), "Roads"]],
                columns=["Type of cost", "Cost USD", "Phase of construction"],
            )
        else:
            equip_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(equip_for_new_roads_cost_usd)
            equipment_costs = pd.DataFrame(
                [["Equipment rental", float(equip_for_new_and_old_roads_cost_usd), "Small DW Roads"]],
                columns=["Type of cost", "Cost USD", "Phase of construction"],
            )

        # add costs for other operations not included in process data for utility mode (e.g., fencing, access roads)
        if calculate_cost_input_dict["turbine_rating_MW"] > 0.1:
            num_turbines = calculate_cost_input_dict["num_turbines"]
            rotor_diameter_m = calculate_cost_input_dict["rotor_diameter_m"]
            construct_duration = calculate_cost_input_dict["construct_duration"]
            num_access_roads = calculate_cost_input_dict["num_access_roads"]

            cost_new_roads_adder = (
                (float(num_turbines) * 17639)
                + (float(num_turbines) * float(rotor_diameter_m) * 24.8)
                + (float(construct_duration) * 55500)
                + float(num_access_roads) * 3800
            )
            cost_adder = self.new_and_existing_total_road_cost(cost_new_roads_adder)
            additional_costs = pd.DataFrame(
                [["Other", float(cost_adder), "Roads"]], columns=["Type of cost", "Cost USD", "Phase of construction"]
            )

        else:  # No 'Other' cost in distributed wind mode:
            cost_new_roads_adder = 0
            cost_adder = self.new_and_existing_total_road_cost(cost_new_roads_adder)
            additional_costs = pd.DataFrame(
                [["Other", float(cost_adder), "Small DW Roads"]],
                columns=["Type of cost", "Cost USD", "Phase of construction"],
            )

        # Create empty road cost (showing cost breakdown by type) dataframe:
        road_cost_columns = ["Type of cost", "Cost USD", "Phase of construction"]

        # Filter out equipment costs from rsmeans tab:
        equipment_data = labor_equip_data[labor_equip_data["Type of cost"] == "Equipment rental"].copy()
        equipment_data["Cost USD"] = (
            equipment_data["Quantity of material"] * equipment_data["Rate USD per unit"]
        ) * calculate_cost_output_dict[
            "wind_multiplier"
        ]  # TODO: Annika can you confirm if this is correct.

        equip_for_new_roads_cost_usd = equipment_data["Cost USD"].sum()
        equip_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(equip_for_new_roads_cost_usd)
        equipment_costs = pd.DataFrame(
            [["Equipment rental", float(equip_for_new_and_old_roads_cost_usd), "Roads"]],
            columns=road_cost_columns,
        )

        # add costs for other operations not included in process data (e.g., fencing, access roads)
        #
        # Assume the following things to compute the cost adder:
        #
        # The crews only work 6 days per week, so a number of days equal to the
        # number of weeks worked is added to convert working days to calendar
        # days.
        #
        # All crews work concurrently, so the total site preparation time is the
        # the max of "Time construct days"

        max_time_construct_days = operation_data["Time construct days"].max()
        num_turbines = float(calculate_cost_input_dict["num_turbines"])
        rotor_diameter_m = float(calculate_cost_input_dict["rotor_diameter_m"])
        num_access_roads = float(calculate_cost_input_dict["num_access_roads"])
        calendar_construct_days = max_time_construct_days + np.ceil(
            max_time_construct_days / 6
        )  # assumes working only 6 days per week
        siteprep_construction_months = calendar_construct_days / 30.0
        cost_new_roads_adder = (
            num_turbines * 17639
            + num_turbines * rotor_diameter_m * 24.8
            + calculate_cost_input_dict["construct_duration"] * 55500
            + num_access_roads * 3800
        )
        cost_adder = self.new_and_existing_total_road_cost(cost_new_roads_adder)
        additional_costs = pd.DataFrame([["Other", cost_adder, "Roads"]], columns=road_cost_columns)

        road_cost = pd.concat((material_costs, equipment_costs, labor_costs, additional_costs))

        # set mobilization cost equal to 5% of total road cost for utility scale model and function of
        # of turbine size for distributed wind:
        if calculate_cost_input_dict["num_turbines"] > 10:
            mobilization_costs_new_roads = road_cost["Cost USD"].sum() * 0.05
            mobilization_costs_new_plus_old_roads = self.new_and_existing_total_road_cost(mobilization_costs_new_roads)
            mobilization_costs = pd.DataFrame(
                [["Mobilization", mobilization_costs_new_plus_old_roads, "Roads"]],
                columns=road_cost_columns,
            )
        else:
            mobilization_costs_new_roads = road_cost["Cost USD"].sum() * self.mobilization_cost_multiplier(
                calculate_cost_input_dict["turbine_rating_MW"]
            )
            mobilization_costs_new_plus_old_roads = self.new_and_existing_total_road_cost(mobilization_costs_new_roads)

            if calculate_cost_input_dict["turbine_rating_MW"] >= 0.1:
                mobilization_costs = pd.DataFrame(
                    [["Mobilization", mobilization_costs_new_plus_old_roads, "Roads"]],
                    columns=road_cost_columns,
                )
            else:
                mobilization_costs = pd.DataFrame(
                    [["Mobilization", mobilization_costs_new_plus_old_roads, "Small DW Roads"]],
                    columns=road_cost_columns,
                )

        road_cost = pd.concat((road_cost, mobilization_costs))
        total_road_cost = road_cost
        calculate_cost_output_dict["total_road_cost"] = total_road_cost
        calculate_cost_output_dict["siteprep_construction_months"] = siteprep_construction_months
        return total_road_cost

    def outputs_for_detailed_tab(self, input_dict, output_dict):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.


        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """
        result = []
        module = type(self).__name__

        # Note that some values are cast with float() so that XlsxWriter
        # (the library depended on by XlsxGenerator) can output them as
        # numbers. XlsxWriter, interestingly, cannot handle np.float32()
        # types.

        result.append(
            {
                "unit": "m^3",
                "type": "variable",
                "variable_df_key_col_name": "Total road volume",
                "value": float(self.output_dict["road_volume_m3"]),
            }
        )

        result.append(
            {
                "unit": "m",
                "type": "variable",
                "variable_df_key_col_name": "Depth to subgrade",
                "value": self.output_dict["depth_to_subgrade_m"],
            }
        )

        result.append(
            {
                "unit": "ft",
                "type": "variable",
                "variable_df_key_col_name": "Crane path width",
                "value": self.output_dict["crane_path_width_m"],  # TODO: Rename variable to: crane_path_width_ft
            }
        )

        if not input_dict["road_distributed_wind"]:
            result.append(
                {
                    "unit": "m",
                    "type": "variable",
                    "variable_df_key_col_name": "Road length",
                    "value": float(self.output_dict["road_length_m"]),
                }
            )

        result.append(
            {
                "unit": "m",
                "type": "variable",
                "variable_df_key_col_name": "Road width",
                "value": self.output_dict["road_width_m"],
            }
        )

        result.append(
            {
                "unit": "m",
                "type": "variable",
                "variable_df_key_col_name": "Road thickness",
                "value": self.output_dict["road_thickness_m"],
            }
        )

        result.append(
            {
                "unit": "cubic yards",
                "type": "variable",
                "variable_df_key_col_name": "Material volume",
                "value": float(self.output_dict["material_volume_cubic_yards"]),
            }
        )

        result.append(
            {
                "unit": "cubic yards",
                "type": "variable",
                "variable_df_key_col_name": "Topsoil volume",
                "value": float(self.output_dict["topsoil_volume"]),
            }
        )

        if input_dict["turbine_rating_MW"] >= 0.1:
            result.append(
                {
                    "unit": "cubic yards",
                    "type": "variable",
                    "variable_df_key_col_name": "Embankment volume crane",
                    "value": float(self.output_dict["embankment_volume_crane"]),
                }
            )

            result.append(
                {
                    "unit": "cubic yards",
                    "type": "variable",
                    "variable_df_key_col_name": "Embankment volume road",
                    "value": float(self.output_dict["embankment_volume_road"]),
                }
            )

            result.append(
                {
                    "unit": "ft^2",
                    "type": "variable",
                    "variable_df_key_col_name": "Rough grading area",
                    "value": float(self.output_dict["rough_grading_area"]),
                }
            )

        for row in self.output_dict["total_road_cost"].itertuples():
            dashed_row = "{} <--> {} <--> {}".format(row[1], row[3], math.ceil(row[2]))
            result.append(
                {
                    "unit": "",
                    "type": "dataframe",
                    "variable_df_key_col_name": "Type of Cost <--> Phase of Construction <--> Cost in USD ",
                    "value": dashed_row,
                    "last_number": row[2],
                }
            )

        for _dict in result:
            _dict["project_id_with_serial"] = self.project_name
            _dict["module"] = module

        self.output_dict["roads_cost_csv"] = result
        return result

    def outputs_for_module_type_operation(self, input_dict, output_dict):
        """
        Outputs dictionaries that are rows for the
        costs_by_module_type_operation

        Returns
        -------
        list
            List of dicts, with each dict representing a row for
            the output.
        """
        result = []
        module = type(self).__name__
        num_turbines = self.input_dict["num_turbines"]

        costs_by_module_type_operation = output_dict["total_road_cost"]
        for _, row in costs_by_module_type_operation.iterrows():
            _dict = dict()
            row = row.to_dict()
            _dict["operation_id"] = row["Phase of construction"]
            _dict["type_of_cost"] = row["Type of cost"]
            _dict["cost"] = row["Cost USD"]
            result.append(_dict)

        for _dict in result:
            _dict["project_id_with_serial"] = self.project_name
            _dict["module"] = module
            _dict["total_or_turbine"] = "total"
            _dict["num_turbines"] = num_turbines

        output_dict["roads_cost_module_type_operation"] = result
        return result

    def run_module(self):
        """
        Runs the SitePreparation module and populates the IO dictionaries with calculated values.

        """
        try:
            self.calculate_road_properties(self.input_dict, self.output_dict)
            operation_data = self.estimate_construction_time(self.input_dict, self.output_dict)

            # pull only global inputs for weather delay from input_dict
            weather_data_keys = ("wind_shear_exponent", "weather_window")

            # specify roads-specific weather delay inputs
            self.weather_input_dict = dict(
                [(i, self.input_dict[i]) for i in self.input_dict if i in set(weather_data_keys)]
            )
            self.weather_input_dict["start_delay_hours"] = (
                0  # assume zero start for when road construction begins (start at beginning of construction time)
            )
            self.weather_input_dict["critical_wind_speed_m_per_s"] = self.input_dict[
                "critical_speed_non_erection_wind_delays_m_per_s"
            ]
            self.weather_input_dict["wind_height_of_interest_m"] = self.input_dict[
                "critical_height_non_erection_wind_delays_m"
            ]

            # compute and specify weather delay mission time for roads
            duration_construction = operation_data["Time construct days"].max(skipna=True)
            operational_hrs_per_day = self.input_dict["hour_day"][self.input_dict["time_construct"]]
            mission_time_hrs = duration_construction * operational_hrs_per_day
            self.weather_input_dict["mission_time_hours"] = mission_time_hrs

            self.calculate_weather_delay(self.weather_input_dict, self.output_dict)
            self.calculate_costs(self.input_dict, self.output_dict)
            self.outputs_for_detailed_tab(self.input_dict, self.output_dict)
            # self.outputs_for_module_type_operation(self.input_dict, self.output_dict)
            self.output_dict["siteprep_module_type_operation"] = self.outputs_for_costs_by_module_type_operation(
                input_df=self.output_dict["total_road_cost"], project_id=self.project_name, total_or_turbine=True
            )
            return 0, 0  # module ran successfully
        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} SitePreparationCost")
            return 1, error  # module did not run successfully



from __future__ import print_function

import sys

import numpy as np
from scipy.optimize import minimize
from scipy.interpolate import PchipInterpolator, Akima1DInterpolator, pchip

from wisdem.commonse.utilities import arc_length
from wisdem.rotorse.geometry_tools.distfunc import distfunc
from wisdem.rotorse.geometry_tools.geom_tools import curvature
from wisdem.rotorse.geometry_tools.cubicspline import NaturalCubicSpline


class Curve(object):
    def __init__(self, points=None):
        super(Curve, self).__init__()

        self.length = 0.0  # Total curve length
        self.s = np.array([])  # Curve accumulated curve length
        self.points = np.array([])  # coordinates of curve
        self.ni = 0  # Number of points

        if points is not None:
            self.initialize(points)

    def initialize(self, points):
        self.points = points
        self.ni = points.shape[0]

        self._compute_s()
        self._compute_dp()
        self._build_splines()

    def _compute_s(self):
        """
        compute normalized curve length
        """
        s = arc_length(self.points)
        self.length = s[-1]
        self.ds = np.diff(s)
        self.s = s / s[-1]

    def _compute_dp(self):
        """compute the unit direction vectors along the curve"""

        t1 = np.gradient(self.points[:, :])[0]
        self.dp = np.array([t1[i, :] / np.linalg.norm(t1[i, :]) for i in range(np.shape(t1)[0])])

    def _build_splines(self):
        self._splines = []

        for j in range(np.shape(self.points)[1]):
            self._splines.append(PchipInterpolator(self.s, self.points[:, j]))

    def redistribute(self, dist=None, s=None):
        if dist is not None:
            self.s = distfunc(dist)
        else:
            self.s = s

        self.ni = np.shape(self.s)[0]
        points = np.zeros((self.ni, np.shape(self.points)[1]))
        for i in range(points.shape[1]):
            points[:, i] = self._splines[i](self.s)

        self.initialize(points)

    def interp_s(self, s):
        """
        interpolate (x,y) at some curve fraction s
        """

        p = np.zeros(np.shape(self.points)[1])
        for i in range(np.shape(self.points)[1]):
            p[i] = self._splines[i](s)

        return p


class AirfoilShape(Curve):
    """
    Base class for airfoil shapes.

    The class automatically computes the LE and TE
    and can redistribute the points smoothly along the surface.
    Points along the surface need to be defined starting at the
    TE pressure side ending at the TE suction side.
    """

    def initialize(self, points):
        self.LE = np.array([])  # Leading edge coordinates
        self.TE = np.array([])  # Trailing edge coordinates
        self.sLE = 0.0  # Leading edge curve fraction
        self.chord = 0.0  # chord length

        super(AirfoilShape, self).initialize(points)
        self.computeLETE()

    def computeLETE(self):
        """
        computes the leading and trailing edge of the airfoil.

        TE is computed as the mid-point between lower and upper TE points
        LE is computed as the point with maximum distance from the TE.
        """

        self.TE = np.array([np.average(self.points[[0, -1], 0]), np.average(self.points[[0, -1], 1])])

        res = minimize(self._sdist, (0.5), method="SLSQP", bounds=[(0, 1)])
        self.sLE = res["x"][0]
        xLE = self._splines[0](self.sLE)
        yLE = self._splines[1](self.sLE)
        self.LE = np.array([xLE, yLE])
        self.curvLE = NaturalCubicSpline(self.s, curvature(self.points))(self.sLE)
        self.chord = np.linalg.norm(self.LE - self.TE)

    def _sdist(self, s):
        x = self._splines[0](s)
        y = self._splines[1](s)
        return -(((x - self.TE[0]) ** 2 + (y - self.TE[1]) ** 2) ** 0.5)

    def leading_edge_dist(self, ni):
        """function that returns a suitable cell size based on airfoil LE curvature"""

        min_ds1 = 1.0 / ni * 0.1
        max_ds1 = 1.0 / ni * 0.5

        ds1 = max((min_ds1 - max_ds1) / 30.0 * abs(self.curvLE) + max_ds1, min_ds1)

        return ds1

    def redistribute(self, ni, even=False, dist=None, dLE=False, dTE=-1.0, s=None):
        """
        redistribute the points on the airfoil using fusedwind.lib.distfunc

        Parameters
        ----------
        ni : int
            total number of points on airfoil
        even : bool
            flag for getting an even distribution of points
        dist : list
            optional list of control points with the form
            [[s0, ds0, n0], [s1, ds1, n1], ... [s<n>, ds<n>, n<n>]]
            where\n
            s<n> is the normalized curve fraction at each control point,\n
            ds<n> is the normalized cell size at each control point,\n
            n<n> is the cell count at each control point.
        dLE : bool
            optional flag for automatically calculating a suitable leading edge cell
            size based on the local curvature
        dTE : float
            optional trailing edge cell size. If set to -1 the cell size will increase
            from the LE to TE according to the tanh distribution function used
            in distfunc
        """

        if even:
            dist = [
                [0, 1.0 / np.float(ni - 1), 1],
                [self.sLE, 1.0 / np.float(ni - 1), int(ni * self.sLE)],
                [1, 1.0 / np.float(ni - 1), ni],
            ]
        elif dLE:
            dist = [[0.0, dTE, 1], [self.sLE, self.leading_edge_dist(ni), ni / 2], [1.0, dTE, ni]]

        super(AirfoilShape, self).redistribute(dist, s=s)

        return self

    def redistribute_chordwise(self, dist):
        """
        redistribute the airfoil according to a chordwise distribution
        """

        # self.redistribute(self.ni, even=True)
        iLE = np.argmin(self.points[:, 0])
        ni = np.shape(dist)[0]
        dist = np.asarray(dist)
        points = np.zeros((np.shape(dist)[0] * 2 - 1, np.shape(self.points)[1]))

        # interpolate pressure side coordinates
        yps = NaturalCubicSpline(self.points[: iLE + 1, 0][::-1], self.points[: iLE + 1, 1][::-1])
        ps = yps(dist)
        # interpolate suction side coordinates
        yss = NaturalCubicSpline(self.points[iLE:, 0], self.points[iLE:, 1])
        ss = yss(dist)
        points[: ni - 1, 0] = dist[::-1][:-1]
        points[ni - 1 :, 0] = dist
        points[:, 1] = np.append(ps[::-1][:-1], ss, axis=0)
        return AirfoilShape(points)

    def s_to_11(self, s):
        """
        Transform the s coordinates from AirfoilShape format:

        * s=0 at TE pressure side (lower surface)
        * s=1 at TE suction side (upper surface)

        to the s coordinates from the input definition:

        * s=0 at LE
        * s=1 at TE suction side (upper surface)
        * s=-1 at TE pressure side (lower surface)
        """

        if s > self.sLE:
            return (s - self.sLE) / (1.0 - self.sLE)
        else:
            return -1.0 + s / self.sLE

    def s_to_01(self, s):
        """
        Transform the s coordinates from the input definition:

        * s=0 at LE
        * s=1 at TE suction side (upper surface)
        * s=-1 at TE pressure side (lower surface)

        to the backend defintion compatible with AirfoilShape():

        * s=0 at TE pressure side (lower surface)
        * s=1 at TE suction side (upper surface)
        """
        if s >= 0.0:
            return s * (1.0 - self.sLE) + self.sLE
        else:
            return (1.0 + s) * self.sLE

    def gurneyflap(self, gf_height, gf_length_factor):
        """add a Gurney flap shaped using a tanh function"""

        if gf_height == 0.0:
            return
        # if the length is not specified it is set to 3 x gf_height
        gf_length = gf_length_factor * gf_height

        # identify starting point of the gf along the chord
        x_gf = 1.0 - gf_length
        id1 = (np.abs(x_gf - self.points[0 : self.ni / 2, 0])).argmin() + 1
        s = np.linspace(x_gf, self.points[0, 0], 100)
        smax = s[-1] - s[0]
        h = np.zeros(100)
        for i in range(100):
            h[i] = (min(0.90 * gf_height, gf_height * (-np.tanh((s[i] - s[0]) / smax * 3) + 1.0))) / 0.90
        h = h[::-1]
        self.gfs = s
        self.gfh = h

        # add the gf shape to the airfoil
        points = self.points.copy()
        for i in range(0, id1):
            points[i, 1] = points[i, 1] - np.interp(points[i, 0], s, h)

        return AirfoilShape(points)

    def open_trailing_edge(self, t):
        """
        add thickness to airfoil
        """

        t0 = np.abs(self.points[-1, 1] - self.points[0, 1])
        dt = (t - t0) / 2.0
        print("dt", dt)
        # linearly add thickness from LE to TE
        iLE = np.argmin(self.points[:, 0])
        xLE = self.points[iLE, 0]
        tlin = np.array([np.linspace(xLE, self.TE[0], 100), np.linspace(0.0, dt, 100)]).T

        tspline = NaturalCubicSpline(tlin[:, 0], tlin[:, 1])

        ys = tspline(self.points[iLE:, 0]) + self.points[iLE:, 1]
        yp = -tspline(self.points[:iLE, 0][::-1])[::-1] + self.points[:iLE, 1]

        self.points[iLE:, 1] = ys
        self.points[:iLE, 1] = yp
        self.initialize(self.points)


def remap2grid(x_ref, y_ref, x, spline=PchipInterpolator, axis=-1):
    try:
        if axis != -1:
            spline_y = spline(x_ref, y_ref, axis=axis)
        else:
            spline_y = spline(x_ref, y_ref)
    except:
        x_ref = np.flip(x_ref, axis=0)
        y_ref = np.flip(y_ref, axis=0)
        if axis != -1:
            spline_y = spline(x_ref, y_ref, axis=axis)
        else:
            spline_y = spline(x_ref, y_ref)

    # error handling for x[-1] - x_ref[-1] > 0 and x[-1]~x_ref[-1]
    try:
        _ = iter(x)
        if x[-1] > max(x_ref) and np.isclose(x[-1], x_ref[-1]):
            x[-1] = x_ref[-1]
    except:
        if np.isclose(x, 0.0):
            x = 0.0
        if x > max(x_ref) and np.isclose(x, x_ref[-1]):
            x = x_ref[-1]

    y_out = spline_y(x)

    np.place(y_out, y_out < np.min(y_ref), np.min(y_ref))
    np.place(y_out, y_out > np.max(y_ref), np.max(y_ref))

    return y_out


def trailing_edge_smoothing(data):
    # correction to trailing edge shape for interpolated airfoils that smooths out unrealistic geometric errors
    # often brought about when transitioning between round, flatback, or sharp trailing edges

    # correct for self cross of TE (rare interpolation error)
    if data[-1, 1] < data[0, 1]:
        temp = data[0, 1]
        data[0, 1] = data[-1, 1]
        data[-1, 1] = temp

    # Find indices on Suction and Pressure side for last 85-95% and 95-100% chordwise
    idx_85_95 = [i_x for i_x, xi in enumerate(data[:, 0]) if xi > 0.85 and xi < 0.95]
    idx_95_100 = [i_x for i_x, xi in enumerate(data[:, 0]) if xi > 0.95 and xi < 1.0]

    idx_85_95_break = [i_idx for i_idx, d_idx in enumerate(np.diff(idx_85_95)) if d_idx > 1][0] + 1
    idx_85_95_SS = idx_85_95[:idx_85_95_break]
    idx_85_95_PS = idx_85_95[idx_85_95_break:]

    idx_95_100_break = [i_idx for i_idx, d_idx in enumerate(np.diff(idx_95_100)) if d_idx > 1][0] + 1
    idx_95_100_SS = idx_95_100[:idx_95_100_break]
    idx_95_100_PS = idx_95_100[idx_95_100_break:]

    # Interpolate the last 5% to the trailing edge
    idx_in_PS = idx_85_95_PS + [-1]
    x_corrected_PS = data[idx_95_100_PS, 0]
    y_corrected_PS = remap2grid(data[idx_in_PS, 0], data[idx_in_PS, 1], x_corrected_PS)

    idx_in_SS = [0] + idx_85_95_SS
    x_corrected_SS = data[idx_95_100_SS, 0]
    y_corrected_SS = remap2grid(data[idx_in_SS, 0], data[idx_in_SS, 1], x_corrected_SS)

    # Overwrite profile with corrected TE
    data[idx_95_100_SS, 1] = y_corrected_SS
    data[idx_95_100_PS, 1] = y_corrected_PS

    return data



import unittest
import os

import numpy as np
import numpy.testing as npt

import wisdem.precomp.properties as prop
import wisdem.precomp.precomp_to_beamdyn as pc2bd
#from wisdem.precomp._precomp import precomp as _precomp

try:
   import cPickle as pickle
except Exception:
   import pickle

mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file
# https://stackoverflow.com/questions/20716812/saving-and-loading-multiple-objects-in-pickle-file
def loadall(filename):
    with open(os.path.join(mydir, filename), "rb") as f:
        while True:
            try:
                yield pickle.load(f)
            except EOFError:
                break

class TestPreCompProperties(unittest.TestCase):
    def test_tw_rate(self):

        # From NREL 5 MW
        myr = np.array([ 1.5,         3.62068966,  5.74137931,  7.86206897,  9.98275862, 12.10344828,
                         14.22413793, 16.34482759, 18.46551724, 20.5862069 , 22.70689655, 24.82758621,
                         26.94827586, 29.06896552, 31.18965517, 33.31034483, 35.43103448, 37.55172414,
                         39.67241379, 41.79310345, 43.9137931 , 46.03448276, 48.15517241, 50.27586207,
                         52.39655172, 54.51724138, 56.63793103, 58.75862069, 60.87931034, 63.        ])
        th0 = np.zeros(myr.shape)
        #th_prime_fort = _precomp.tw_rate(myr, th0)
        th_prime_fort = th0
        th_prime_py = prop.tw_rate(myr, th0)
        npt.assert_almost_equal(th_prime_fort, th_prime_py)

        th0 = np.linspace(10., -10., myr.size)
        #th_prime_fort = _precomp.tw_rate(myr, th0)
        th_prime_fort = -0.32520325*np.ones(th0.shape)
        th_prime_py = prop.tw_rate(myr, th0)
        npt.assert_almost_equal(th_prime_fort, th_prime_py)

    def test_properties(self):
        fnames = ['section_dump_nrel5mw.pkl', 'section_dump_iea15mw.pkl']
        
        for f in fnames:
            with self.subTest(f=f):
                myitems = loadall(f)
                nsec = myitems.__next__()
                for k in range(nsec):
                    with self.subTest(i=k):
                        chord = myitems.__next__()
                        theta = myitems.__next__()
                        th_prime = myitems.__next__()
                        le_loc = myitems.__next__()
                        xnode = myitems.__next__()
                        ynode = myitems.__next__()
                        E1 = myitems.__next__()
                        E2 = myitems.__next__()
                        G12 = myitems.__next__()
                        nu12 = myitems.__next__()
                        rho = myitems.__next__()
                        locU = myitems.__next__()
                        n_laminaU = myitems.__next__()
                        n_pliesU = myitems.__next__()
                        tU = myitems.__next__()
                        thetaU = myitems.__next__()
                        mat_idxU = myitems.__next__()
                        locL = myitems.__next__()
                        n_laminaL = myitems.__next__()
                        n_pliesL = myitems.__next__()
                        tL = myitems.__next__()
                        thetaL = myitems.__next__()
                        mat_idxL = myitems.__next__()
                        nwebs = myitems.__next__()
                        locW = myitems.__next__()
                        n_laminaW = myitems.__next__()
                        n_pliesW = myitems.__next__()
                        tW = myitems.__next__()
                        thetaW = myitems.__next__()
                        mat_idxW = myitems.__next__()

                        results_fort = myitems.__next__()
                        results_py = prop.properties(chord,
                                                     theta,
                                                     th_prime,
                                                     le_loc,
                                                     xnode,
                                                     ynode,
                                                     E1,
                                                     E2,
                                                     G12,
                                                     nu12,
                                                     rho,
                                                     locU,
                                                     n_laminaU,
                                                     n_pliesU,
                                                     tU,
                                                     thetaU,
                                                     mat_idxU,
                                                     locL,
                                                     n_laminaL,
                                                     n_pliesL,
                                                     tL,
                                                     thetaL,
                                                     mat_idxL,
                                                     nwebs,
                                                     locW,
                                                     n_laminaW,
                                                     n_pliesW,
                                                     tW,
                                                     thetaW,
                                                     mat_idxW,
                                                     )

                        npt.assert_almost_equal(results_fort, results_py, decimal=5)

    def test_match_anba(self):

        # Stiffness and inertia matrices from https://github.com/WISDEM/SONATA/tree/develop/examples/1_IEA15MW
        # for the spanwise section 0.517241 
        K_anba = np.array([
            [1.0715521486176741e+08, 3.5706427831706675e+06, 0.0000000000000000e+00, 0.0000000000000000e+00, 0.0000000000000000e+00, 8.7833838441895731e+06],
            [3.5706427832182581e+06, 3.0025188389417487e+08, 0.0000000000000000e+00, 0.0000000000000000e+00, 0.0000000000000000e+00, 1.2073068428051447e+07],
            [0.0000000000000000e+00, 0.0000000000000000e+00, 1.9613892805330948e+10, 5.4105970862822247e+09, -2.0501694138219514e+08, 0.0000000000000000e+00],
            [0.0000000000000000e+00, 0.0000000000000000e+00, 5.4105970862821531e+09, 1.4137313098254759e+10, -2.3555582131167597e+08, 0.0000000000000000e+00],
            [0.0000000000000000e+00, 0.0000000000000000e+00, -2.0501694138206977e+08, -2.3555582131167600e+08, 4.3043864825780182e+09, 0.0000000000000000e+00],
            [8.7833838441405557e+06, 1.2073068427973984e+07, 0.0000000000000000e+00, 0.0000000000000000e+00, 0.0000000000000000e+00, 1.9466974220949066e+08],
        ])
        

        I_anba = np.array([
            [3.6479516547101201e+02, 0.0000000000000000e+00, 0.0000000000000000e+00, 0.0000000000000000e+00, 0.0000000000000000e+00, -2.5786979371895751e+02],
            [0.0000000000000000e+00, 3.6479516547101201e+02, 0.0000000000000000e+00, 0.0000000000000000e+00, 0.0000000000000000e+00, 8.2707311243991697e+00],
            [0.0000000000000000e+00, 0.0000000000000000e+00, 3.6479516547101201e+02, 2.5786979371895751e+02, -8.2707311243991697e+00, 0.0000000000000000e+00],
            [0.0000000000000000e+00, 0.0000000000000000e+00, 2.5786979371895751e+02, 6.7555496585520257e+02, -1.5049443526260998e+01, 0.0000000000000000e+00],
            [0.0000000000000000e+00, 0.0000000000000000e+00, -8.2707311243991697e+00, -1.5049443526260998e+01, 5.8584394201929328e+01, 0.0000000000000000e+00],
            [-2.5786979371895751e+02, 8.2707311243991697e+00, 0.0000000000000000e+00, 0.0000000000000000e+00, 0.0000000000000000e+00, 7.3413936005712992e+02],
        ])
        
        
        
        fnames = ['section_dump_iea15mw.pkl']
        
        for f in fnames:
            with self.subTest(f=f):
                myitems = loadall(f)
                nsec = myitems.__next__()
                for k in range(nsec):
                    with self.subTest(i=k):
                        chord = myitems.__next__()
                        theta = myitems.__next__()
                        th_prime = myitems.__next__()
                        le_loc = myitems.__next__()
                        xnode = myitems.__next__()
                        ynode = myitems.__next__()
                        E1 = myitems.__next__()
                        E2 = myitems.__next__()
                        G12 = myitems.__next__()
                        nu12 = myitems.__next__()
                        rho = myitems.__next__()
                        locU = myitems.__next__()
                        n_laminaU = myitems.__next__()
                        n_pliesU = myitems.__next__()
                        tU = myitems.__next__()
                        thetaU = myitems.__next__()
                        mat_idxU = myitems.__next__()
                        locL = myitems.__next__()
                        n_laminaL = myitems.__next__()
                        n_pliesL = myitems.__next__()
                        tL = myitems.__next__()
                        thetaL = myitems.__next__()
                        mat_idxL = myitems.__next__()
                        nwebs = myitems.__next__()
                        locW = myitems.__next__()
                        n_laminaW = myitems.__next__()
                        n_pliesW = myitems.__next__()
                        tW = myitems.__next__()
                        thetaW = myitems.__next__()
                        mat_idxW = myitems.__next__()

                        results_fort = myitems.__next__()
                        results_py = prop.properties(chord,
                                                     theta,
                                                     th_prime,
                                                     le_loc,
                                                     xnode,
                                                     ynode,
                                                     E1,
                                                     E2,
                                                     G12,
                                                     nu12,
                                                     rho,
                                                     locU,
                                                     n_laminaU,
                                                     n_pliesU,
                                                     tU,
                                                     thetaU,
                                                     mat_idxU,
                                                     locL,
                                                     n_laminaL,
                                                     n_pliesL,
                                                     tL,
                                                     thetaL,
                                                     mat_idxL,
                                                     nwebs,
                                                     locW,
                                                     n_laminaW,
                                                     n_pliesW,
                                                     tW,
                                                     thetaW,
                                                     mat_idxW,
                                                     )

                    if k==15:
                        (eifbar,eilbar,gjbar,eabar,eiflbar,
                        sfbar,slbar,sftbar,sltbar,satbar,
                        z_sc,y_sc,ztc_ref,ytc_ref,
                        mass,area,iflap_eta,ilag_zeta,tw_iner,
                        zcm_ref,ycm_ref) = results_py

                        EIxx  = eilbar
                        EIyy  = eifbar
                        GJ  = gjbar
                        EA  = eabar
                        EIxy  = eiflbar
                        EA_EIxx  = slbar
                        EA_EIyy  = sfbar
                        EIxx_GJ  = sltbar
                        EIyy_GJ  = sftbar
                        EA_GJ  = satbar
                        x_sc  = z_sc
                        y_sc  = y_sc
                        x_tc  = ztc_ref
                        y_tc  = ytc_ref
                        rhoA  = mass
                        A  = area
                        flap_iner  = iflap_eta
                        edge_iner  = ilag_zeta
                        Tw_iner  = tw_iner
                        x_cg  = zcm_ref
                        y_cg  = ycm_ref

                        K_precomp = np.zeros((6,6))
                        I_precomp = np.zeros((6,6))

                        # Build stiffness matrix at the reference axis
                        K_precomp = pc2bd.pc2bd_K(
                            EA,
                            EIxx,
                            EIyy,
                            EIxy,
                            EA_EIxx,
                            EA_EIyy,
                            EIxx_GJ,
                            EIyy_GJ,
                            EA_GJ,
                            GJ,
                            flap_iner+edge_iner,
                            edge_iner,
                            flap_iner,
                            x_sc,
                            y_sc,
                            )
                        # Build inertia matrix at the reference axis
                        I_precomp = pc2bd.pc2bd_I(
                            rhoA,
                            edge_iner,
                            flap_iner,
                            edge_iner+flap_iner,
                            x_cg,
                            y_cg,
                            np.deg2rad(Tw_iner),
                            np.deg2rad(theta),
                            )
                        # Only check diagonal elements of K and I
                        ix = np.arange(6)
                        max_pc_error_K = np.array([25, 25, 11, 16, 11, 5]) # Relative errors at the time of setting up test
                        rel_err_K = np.zeros(len(ix))
                        for i in range(len(ix)):
                            rel_err_K[i] = abs((K_precomp[ix[i],ix[i]]-K_anba[ix[i],ix[i]])/K_anba[ix[i],ix[i]]*100.)
                        npt.assert_array_less(rel_err_K, max_pc_error_K)
                        
                        # Check all elements of I
                        max_pc_error_I = np.array([9, 9, 9, 12, 12, 12]) # Relative errors at the time of setting up test. Updated after switching rotation and translation
                        rel_err_I = np.zeros(len(ix))
                        for i in range(len(ix)):
                            rel_err_I[i] = abs((I_precomp[ix[i],ix[i]]-I_anba[ix[i],ix[i]])/I_anba[ix[i],ix[i]]*100.)
                        npt.assert_array_less(rel_err_I, max_pc_error_I)




if __name__ == "__main__":
    unittest.main()



import os

from wisdem import run_wisdem

## File management
mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file
fname_wt_input = mydir + os.sep + "IEA-10-198-RWT.yaml"
fname_modeling_options = mydir + os.sep + "modeling_options_iea10.yaml"
fname_analysis_options = mydir + os.sep + "analysis_options.yaml"

wt_opt, analysis_options, opt_options = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)



"""
This file postprocesses results from WISDEM. This script is installed
as a console script when WISDEM is installed.

This script allows you to compare wind turbine designs via the
outputted yaml files. This script runs an instance of WISDEM for each yaml
file then produces plots and text output comparing the designs and performance.
You can modify the yaml files being compared, as well as which values are plotted
and printed to screen.
See the bottom of this file for some user customization options.

At a terminal, example usage is:

compare_designs blade.yaml optimized_blade.yaml

This will print results to screen, then create and save plots.
"""

import os
import sys
import argparse

import numpy as np
import matplotlib.pyplot as plt

import wisdem.postprocessing.wisdem_get as getter
from wisdem.glue_code.runWISDEM import run_wisdem, load_wisdem

this_dir = os.path.dirname(os.path.realpath(__file__))

# These are the modeling and analysis options used to evaluate the designs.
# Both of these yaml files are used for all yamls to make a fair comparison.
fname_modeling_options_default = this_dir + os.sep + "default_modeling_options.yaml"
fname_analysis_options_default = this_dir + os.sep + "default_analysis_options.yaml"


def create_all_plots(
    list_of_sims,
    list_of_labels,
    modeling_options,
    analysis_options,
    folder_output,
    show_plots,
    font_size,
    extension,
):
    mult_flag = len(list_of_sims) > 1

    colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
    if not mult_flag:
        colors = ["k"] + colors

    # Twist
    try:
        ftw, axtw = plt.subplots(1, 1, figsize=(5.3, 4))

        for idx, (yaml_data, label) in enumerate(zip(list_of_sims, list_of_labels)):
            s_opt_twist = yaml_data["blade.opt_var.s_opt_twist"]
            twist_opt = yaml_data["blade.opt_var.twist_opt"]
            axtw.plot(
                yaml_data["blade.outer_shape_bem.s"],
                np.rad2deg(yaml_data["rotorse.theta"]),
                "-",
                color=colors[idx],
                label=label,
            )
            axtw.plot(s_opt_twist, np.rad2deg(twist_opt), "o", color=colors[idx], markersize=3)

        s_opt_twist = list_of_sims[0]["blade.outer_shape_bem.s"]
        twist_opt = list_of_sims[0]["blade.pa.twist_param"]
        axtw.plot(
            s_opt_twist,
            np.rad2deg(
                twist_opt - analysis_options["design_variables"]["blade"]["aero_shape"]["twist"]["max_decrease"]
            ),
            ":o",
            color=colors[idx + 1],
            markersize=3,
            label="Bounds",
        )
        axtw.plot(
            s_opt_twist,
            np.rad2deg(
                twist_opt + analysis_options["design_variables"]["blade"]["aero_shape"]["twist"]["max_increase"]
            ),
            ":o",
            color=colors[idx + 1],
            markersize=3,
        )

        if mult_flag:
            axtw.legend(fontsize=font_size)

        plt.xlabel("Blade Nondimensional Span [-]", fontsize=font_size + 2, fontweight="bold")
        plt.ylabel("Twist [deg]", fontsize=font_size + 2, fontweight="bold")
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")
        plt.subplots_adjust(bottom=0.15, left=0.15)
        fig_name = "twist_opt" + extension
        ftw.savefig(os.path.join(folder_output, fig_name), pad_inches=0.1, bbox_inches="tight")
        plt.close()
    except KeyError:
        pass

    try:
        # Chord
        fc, axc = plt.subplots(1, 1, figsize=(5.3, 4))

        for idx, (yaml_data, label) in enumerate(zip(list_of_sims, list_of_labels)):
            s_opt_chord = yaml_data["blade.opt_var.s_opt_chord"]
            chord_opt = yaml_data["blade.opt_var.chord_opt"]
            axc.plot(
                yaml_data["blade.outer_shape_bem.s"],
                yaml_data["blade.pa.chord_param"],
                "-",
                color=colors[idx],
                label=label,
            )
            axc.plot(s_opt_chord, chord_opt, "o", color=colors[idx], markersize=3)

        s_opt_chord = list_of_sims[0]["blade.outer_shape_bem.s"]
        chord_opt = list_of_sims[0]["blade.pa.chord_param"]
        axc.plot(
            s_opt_chord,
            np.array(analysis_options["design_variables"]["blade"]["aero_shape"]["chord"]["max_decrease"]) * chord_opt,
            ":o",
            color=colors[idx + 1],
            markersize=3,
            label="Bounds",
        )
        axc.plot(
            s_opt_chord,
            np.array(analysis_options["design_variables"]["blade"]["aero_shape"]["chord"]["max_increase"]) * chord_opt,
            ":o",
            color=colors[idx + 1],
            markersize=3,
        )

        if mult_flag:
            axc.legend(fontsize=font_size)
        plt.xlabel("Blade Nondimensional Span [-]", fontsize=font_size + 2, fontweight="bold")
        plt.ylabel("Chord [m]", fontsize=font_size + 2, fontweight="bold")
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")
        plt.subplots_adjust(bottom=0.15, left=0.15)
        fig_name = "chord" + extension
        fc.savefig(os.path.join(folder_output, fig_name), pad_inches=0.1, bbox_inches="tight")
        plt.close()
    except KeyError:
        pass

    # Struct Layers Blade
    try:
        n_layers = modeling_options["WISDEM"]["RotorSE"]["n_layers"]
        layer_name = modeling_options["WISDEM"]["RotorSE"]["layer_name"]
        for i in range(n_layers):
            f, ax = plt.subplots(1, 1, figsize=(5.3, 4))
            for idx, (yaml_data, label) in enumerate(zip(list_of_sims, list_of_labels)):
                ax.plot(
                    yaml_data["blade.outer_shape_bem.s"],
                    yaml_data["blade.ps.layer_thickness_param"][i, :] * 1e3,
                    "-",
                    color=colors[idx],
                    label=label + ' %s'%layer_name[i],
                )
            if mult_flag:
                ax.legend(fontsize=font_size)
            plt.xlabel("Blade Nondimensional Span [-]", fontsize=font_size + 2, fontweight="bold")
            plt.ylabel("Layer Thickness [mm]", fontsize=font_size + 2, fontweight="bold")
            plt.xticks(fontsize=font_size)
            plt.yticks(fontsize=font_size)
            plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")
            plt.subplots_adjust(bottom=0.15, left=0.15)
            fig_name = "blade_layer_%d_thick"%i + extension
            f.savefig(os.path.join(folder_output, fig_name), pad_inches=0.1, bbox_inches="tight")
            plt.close()
    except KeyError:
        pass

    # Strains spar caps
    try:
        feps, axeps = plt.subplots(1, 1, figsize=(5.3, 4))
        for idx, (yaml_data, label) in enumerate(zip(list_of_sims, list_of_labels)):
            axeps.plot(
                yaml_data["blade.outer_shape_bem.s"],
                yaml_data["rotorse.rs.strains.strainU_spar"] * 1.0e6,
                "-",
                color=colors[idx],
                label=label,
            )
            axeps.plot(
                yaml_data["blade.outer_shape_bem.s"],
                yaml_data["rotorse.rs.strains.strainL_spar"] * 1.0e6,
                "-",
                color=colors[idx],
            )

        plt.ylim([-5e3, 5e3])
        if mult_flag:
            axeps.legend(fontsize=font_size)
        plt.xlabel("Blade Nondimensional Span [-]", fontsize=font_size + 2, fontweight="bold")
        plt.ylabel("Spar Caps Strains [mu eps]", fontsize=font_size + 2, fontweight="bold")
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")
        plt.subplots_adjust(bottom=0.15, left=0.2)
        fig_name = "strains_sc_opt" + extension
        feps.savefig(os.path.join(folder_output, fig_name), pad_inches=0.1, bbox_inches="tight")
        plt.close()
    except KeyError:
        pass

    # Strains trailing edge
    try:
        fete, axete = plt.subplots(1, 1, figsize=(5.3, 4))
        for idx, (yaml_data, label) in enumerate(zip(list_of_sims, list_of_labels)):
            axete.plot(
                yaml_data["blade.outer_shape_bem.s"],
                yaml_data["rotorse.rs.frame.strainU_te"] * 1.0e6,
                "-",
                color=colors[idx],
                label=label,
            )
            axete.plot(
                yaml_data["blade.outer_shape_bem.s"],
                yaml_data["rotorse.rs.frame.strainL_te"] * 1.0e6,
                "-",
                color=colors[idx],
            )

        plt.ylim([-5e3, 5e3])
        if mult_flag:
            axete.legend(fontsize=font_size)
        plt.xlabel("Blade Nondimensional Span [-]", fontsize=font_size + 2, fontweight="bold")
        plt.ylabel("Trailing Edge Strains [mu eps]", fontsize=font_size + 2, fontweight="bold")
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")
        plt.subplots_adjust(bottom=0.15, left=0.2)
        fig_name = "strains_te_opt" + extension
        fete.savefig(os.path.join(folder_output, fig_name), pad_inches=0.1, bbox_inches="tight")
        plt.close()
    except KeyError:
        pass

    # Angle of attack and stall angle
    try:
        faoa, axaoa = plt.subplots(1, 1, figsize=(5.3, 4))
        for idx, (yaml_data, label) in enumerate(zip(list_of_sims, list_of_labels)):
            axaoa.plot(
                yaml_data["rotorse.s"],
                yaml_data["rotorse.stall_check.aoa_along_span"],
                "-",
                color=colors[idx],
                label=label,
            )
        axaoa.plot(
            yaml_data["rotorse.s"],
            yaml_data["rotorse.stall_check.stall_angle_along_span"],
            ":",
            color=colors[idx + 1],
            label="Stall",
        )
        if mult_flag:
            axaoa.legend(fontsize=font_size)
        axaoa.set_ylim([0, 20])
        plt.xlabel("Blade Nondimensional Span [-]", fontsize=font_size + 2, fontweight="bold")
        plt.ylabel("Angle of Attack [deg]", fontsize=font_size + 2, fontweight="bold")
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")
        plt.subplots_adjust(bottom=0.15, left=0.15)
        fig_name = "aoa" + extension
        faoa.savefig(os.path.join(folder_output, fig_name), pad_inches=0.1, bbox_inches="tight")
        plt.close()
    except KeyError:
        pass

    # Airfoil efficiency
    try:
        feff, axeff = plt.subplots(1, 1, figsize=(5.3, 4))
        for idx, (yaml_data, label) in enumerate(zip(list_of_sims, list_of_labels)):
            axeff.plot(
                yaml_data["blade.outer_shape_bem.s"],
                yaml_data["rotorse.rp.powercurve.cl_regII"] / yaml_data["rotorse.rp.powercurve.cd_regII"],
                "-",
                color=colors[idx],
                label=label,
            )
        if mult_flag:
            axeff.legend(fontsize=font_size)
        plt.xlabel("Blade Nondimensional Span [-]", fontsize=font_size + 2, fontweight="bold")
        plt.ylabel("Airfoil Efficiency [-]", fontsize=font_size + 2, fontweight="bold")
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")
        plt.subplots_adjust(bottom=0.15, left=0.15)
        fig_name = "af_efficiency" + extension
        feff.savefig(os.path.join(folder_output, fig_name), pad_inches=0.1, bbox_inches="tight")
        plt.close()
    except KeyError:
        pass

    # Prebend and sweep
    try:
        fbend, axbend = plt.subplots(1, 1, figsize=(5.3, 4))
        for idx, (yaml_data, label) in enumerate(zip(list_of_sims, list_of_labels)):
            axbend.plot(
                yaml_data["blade.outer_shape_bem.s"],
                yaml_data["blade.outer_shape_bem.ref_axis"][:, 0],
                "-",
                color=colors[idx],
                label=label,
            )
        if mult_flag:
            axbend.legend(fontsize=font_size)
        plt.xlabel("Blade Nondimensional Span [-]", fontsize=font_size + 2, fontweight="bold")
        plt.ylabel("Prebend [m]", fontsize=font_size + 2, fontweight="bold")
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")
        plt.subplots_adjust(bottom=0.15, left=0.15)
        fig_name = "prebend" + extension
        fbend.savefig(os.path.join(folder_output, fig_name), pad_inches=0.1, bbox_inches="tight")
        plt.close()

        fsweep, axsweep = plt.subplots(1, 1, figsize=(5.3, 4))
        for idx, (yaml_data, label) in enumerate(zip(list_of_sims, list_of_labels)):
            axsweep.plot(
                yaml_data["blade.outer_shape_bem.s"],
                yaml_data["blade.outer_shape_bem.ref_axis"][:, 1],
                "-",
                color=colors[idx],
                label=label,
            )
        if mult_flag:
            axsweep.legend(fontsize=font_size)
        plt.xlabel("Blade Nondimensional Span [-]", fontsize=font_size + 2, fontweight="bold")
        plt.ylabel("Sweep [m]", fontsize=font_size + 2, fontweight="bold")
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")
        plt.subplots_adjust(bottom=0.15, left=0.15)
        fig_name = "sweep" + extension
        fsweep.savefig(os.path.join(folder_output, fig_name), pad_inches=0.1, bbox_inches="tight")
        plt.close()
    except KeyError:
        pass

    # Tower geometry
    # try:
    brown = np.array([150.0, 75.0, 0.0]) / 256.0
    ftow = plt.figure(figsize=(11, 4))
    ax1 = ftow.add_subplot(121)
    for idx, (yaml_data, label) in enumerate(zip(list_of_sims, list_of_labels)):
        ax1.plot(
            getter.get_tower_diameter(yaml_data),
            getter.get_zpts(yaml_data),
            "-",
            color=colors[idx],
            label=label,
        )
    vx = ax1.get_xlim()
    zs = getter.get_zpts(list_of_sims[0])
    if zs.min() < -5.0:
        water_depth = list_of_sims[0]["env.water_depth"]
        h_trans = getter.get_transition_height(list_of_sims[0])
        ax1.plot(vx, np.zeros(2), color="b", linestyle="--")
        ax1.plot(vx, -water_depth * np.ones(2), color=brown, linestyle="--")
        ax1.plot(vx, h_trans * np.ones(2), color="g", linestyle="--")
        ax1.text(vx[0] + 0.02 * np.diff(vx), 2, "Water line", color="b", fontsize=12)
        ax1.text(vx[0] + 0.02 * np.diff(vx), -water_depth + 2, "Mud line", color=brown, fontsize=12)
        ax1.text(vx[0] + 0.02 * np.diff(vx), h_trans + 2, "Tower transition", color="g", fontsize=12)
    ax1.set_xlim(vx)
    plt.xlabel("Outer Diameter [m]", fontsize=font_size + 2, fontweight="bold")
    plt.ylabel("Tower Height [m]", fontsize=font_size + 2, fontweight="bold")
    plt.xticks(fontsize=font_size)
    plt.yticks(fontsize=font_size)
    plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")

    ax2 = ftow.add_subplot(122)
    for idx, (yaml_data, label) in enumerate(zip(list_of_sims, list_of_labels)):
        y = 1e3 * getter.get_tower_thickness(yaml_data)
        ax2.step(
            np.r_[y, y[-1]],
            getter.get_zpts(yaml_data),
            "-",
            color=colors[idx],
            label=label,
            where="post",
        )
    vx = ax2.get_xlim()
    if zs.min() < -5.0:
        ax2.plot(vx, np.zeros(2), color="b", linestyle="--")
        ax2.plot(vx, -water_depth * np.ones(2), color=brown, linestyle="--")
        ax2.plot(vx, h_trans * np.ones(2), color="g", linestyle="--")
    ax2.set_xlim(vx)
    if mult_flag:
        ax2.legend(fontsize=font_size)
    plt.xlabel("Wall Thickness [mm]", fontsize=font_size + 2, fontweight="bold")
    plt.xticks(fontsize=font_size)
    plt.setp(ax2.get_yticklabels(), visible=False)
    plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")
    plt.subplots_adjust(bottom=0.15, left=0.15)
    fig_name = "tower-monopile_geometry" + extension
    ftow.subplots_adjust(hspace=0.02, wspace=0.02, bottom=0.15, left=0.15)
    ftow.savefig(os.path.join(folder_output, fig_name), pad_inches=0.1, bbox_inches="tight")
    plt.close()
    # except KeyError:
    #    pass

    def simple_plot_results(x_axis_label, y_axis_label, x_axis_data_name, y_axis_data_name, plot_filename):
        f, ax = plt.subplots(1, 1, figsize=(5.3, 4))
        for i_yaml, yaml_data in enumerate(list_of_sims):
            ax.plot(
                yaml_data[x_axis_data_name],
                yaml_data[y_axis_data_name],
                "-",
                color=colors[i_yaml],
                label=list_of_labels[i_yaml],
            )
        if mult_flag:
            ax.legend(fontsize=font_size)
        plt.xlabel(x_axis_label, fontsize=font_size + 2, fontweight="bold")
        plt.ylabel(y_axis_label, fontsize=font_size + 2, fontweight="bold")
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.grid(color=[0.8, 0.8, 0.8], linestyle="--")
        plt.subplots_adjust(bottom=0.15, left=0.15)
        fig_name = plot_filename + extension
        f.savefig(os.path.join(folder_output, fig_name), pad_inches=0.1, bbox_inches="tight")
        plt.close()

    try:
        # Edgewise stiffness
        simple_plot_results(
            "Blade Nondimensional Span [-]",
            "Edgewise Stiffness [Nm2]",
            "blade.outer_shape_bem.s",
            "rotorse.EIxx",
            "edge",
        )

        # Torsional stiffness
        simple_plot_results(
            "Blade Nondimensional Span [-]",
            "Torsional Stiffness [Nm2]",
            "blade.outer_shape_bem.s",
            "rotorse.GJ",
            "torsion",
        )

        # Flapwise stiffness
        simple_plot_results(
            "Blade Nondimensional Span [-]",
            "Flapwise Stiffness [Nm2]",
            "blade.outer_shape_bem.s",
            "rotorse.EIyy",
            "flap",
        )

        # Mass
        simple_plot_results(
            "Blade Nondimensional Span [-]",
            "Unit Mass [kg/m]",
            "blade.outer_shape_bem.s",
            "rotorse.rhoA",
            "mass",
        )
    except KeyError:
        pass

    try:
        # Relative thickness
        simple_plot_results(
            "Blade Nondimensional Span [-]",
            "Relative Thickness [-]",
            "blade.outer_shape_bem.s",
            "blade.interp_airfoils.r_thick_interp",
            "r_thick",
        )
    except KeyError:
        pass

    try:
        # Induction
        simple_plot_results(
            "Blade Nondimensional Span [-]",
            "Axial Induction [-]",
            "blade.outer_shape_bem.s",
            "rotorse.rp.powercurve.ax_induct_regII",
            "induction",
        )

        # Lift coefficient
        simple_plot_results(
            "Blade Nondimensional Span [-]",
            "Lift Coefficient [-]",
            "blade.outer_shape_bem.s",
            "rotorse.rp.powercurve.cl_regII",
            "lift_coeff",
        )

        # Drag coefficient
        simple_plot_results(
            "Blade Nondimensional Span [-]",
            "Drag Coefficient [-]",
            "blade.outer_shape_bem.s",
            "rotorse.rp.powercurve.cd_regII",
            "drag_coeff",
        )

        # Power curve pitch
        simple_plot_results(
            "Wind velocity [m/s]",
            "Pitch angle [deg]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.pitch",
            "pitch",
        )

        # Power curve power
        simple_plot_results(
            "Wind velocity [m/s]",
            "Electrical Power [W]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.P",
            "power_elec",
        )

        # Power curve power (mechanical)
        simple_plot_results(
            "Wind velocity [m/s]",
            "Mechanical Power [W]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.P_aero",
            "power_aero",
        )

        # Power curve power coeff
        simple_plot_results(
            "Wind velocity [m/s]",
            "Electrical Power Coefficient [-]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.Cp",
            "cp_elec",
        )

        # Power curve power coeff (mechanical)
        simple_plot_results(
            "Wind velocity [m/s]",
            "Mechanical Power Coefficient [-]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.Cp_aero",
            "cp_aero",
        )

        # Power curve rpm
        simple_plot_results(
            "Wind velocity [m/s]",
            "Rotor speed [rpm]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.Omega",
            "omega",
        )

        # Power curve thrust
        simple_plot_results(
            "Wind velocity [m/s]",
            "Thrust [N]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.T",
            "thrust",
        )

        # Power curve thrust coeff
        simple_plot_results(
            "Wind velocity [m/s]",
            "Thrust Coefficient [-]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.Ct_aero",
            "ct_aero",
        )

        # Power curve torque
        simple_plot_results(
            "Wind velocity [m/s]",
            "Torque [Nm]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.Q",
            "torque",
        )

        # Power curve torque coeff
        simple_plot_results(
            "Wind velocity [m/s]",
            "Torque Coefficient [-]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.Cq_aero",
            "cq_aero",
        )

        # Power curve moment
        simple_plot_results(
            "Wind velocity [m/s]",
            "Blade moment [Nm]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.M",
            "moment",
        )

        # Power curve moment coeff
        simple_plot_results(
            "Wind velocity [m/s]",
            "Blade moment coefficient [-]",
            "rotorse.rp.powercurve.V",
            "rotorse.rp.powercurve.Cm_aero",
            "cm_aero",
        )
    except KeyError:
        pass

    if show_plots:
        plt.show()


def print_results_to_screen(list_of_sims, list_of_labels, values_to_print):
    list_of_augmented_labels = []
    max_label_length = 1
    for label in list_of_labels:
        list_of_augmented_labels.append(f"{label:15.15}")

    case_headers = "| Data name       | " + " | ".join(list_of_augmented_labels) + " | Units          |"
    # Header describing what we are printing:
    title_string = "Comparison between WISDEM results from yaml files"
    spacing = (len(case_headers) - len(title_string) - 2) // 2

    print("+" + "-" * (len(case_headers) - 2) + "+")
    print("|" + " " * spacing + title_string + " " * spacing + " |")
    print("+" + "-" * (len(case_headers) - 2) + "+")

    print(case_headers)
    print("+" + "-" * (len(case_headers) - 2) + "+")

    for key in values_to_print:
        value_name = values_to_print[key][0]
        units = values_to_print[key][1]
        units_str = f" | {units}" + (15 - len(str(units))) * " " + "|"

        try:
            value_sizer = list_of_sims[0].get_val(value_name, units)
        except KeyError:
            continue
        size_of_variable = len(value_sizer)

        for idx in range(size_of_variable):
            if size_of_variable > 1:
                augmented_key = f"{key}_{idx}"
            else:
                augmented_key = key

            name_str = f"| {augmented_key}" + (16 - len(augmented_key)) * " " + "| "

            list_of_values = []
            for yaml_data in list_of_sims:
                value = yaml_data.get_val(value_name, units).copy()

                if len(values_to_print[key]) > 2:
                    value *= values_to_print[key][2]

                list_of_values.append(f"{float(value[idx]):15.5f}")

            values_str = " | ".join(list_of_values)
            print(name_str + values_str + units_str)

    print("+" + "-" * (len(case_headers) - 2) + "+")
    print()


def save_lcoe_data_to_file(list_of_sims, folder_output):
    lcoe_data = np.zeros((8, len(list_of_sims)))
    try:
        for idx, yaml_data in enumerate(list_of_sims):
            lcoe_data[0, idx] = yaml_data["financese.turbine_number"]
            lcoe_data[1, idx] = yaml_data["financese.machine_rating"][0]
            lcoe_data[2, idx] = yaml_data["financese.tcc_per_kW"][0]
            lcoe_data[3, idx] = yaml_data["financese.bos_per_kW"][0]
            lcoe_data[4, idx] = yaml_data["financese.opex_per_kW"][0]
            lcoe_data[5, idx] = yaml_data["financese.turbine_aep"][0] * 1.0e-6
            lcoe_data[6, idx] = yaml_data["financese.fixed_charge_rate"][0] * 100.0
            lcoe_data[7, idx] = yaml_data["financese.lcoe"][0] * 1000.0

        np.savetxt(os.path.join(folder_output, "lcoe.dat"), lcoe_data)
    except:
        pass


def run(list_of_sims, list_of_labels, modeling_options, analysis_options):
    # These are options for the plotting and saving
    show_plots = False  # if True, print plots to screen in addition to saving files
    font_size = 12
    extension = ".png"  # '.pdf'
    folder_output = "outputs"

    # These are the values to print to screen for text-based output.
    # The dictionary keys are the value names.
    # The first string in the list is where that value exists in the WISDEM problem,
    # the second string is the units to print the value in,
    # and the optional third string is the multiplicative scalar on the value to be printed.
    values_to_print = {
        "Rotor Diameter": ["blade.high_level_blade_props.rotor_diameter", "m"],
        "TSR": ["control.rated_TSR", None],
        "AEP": ["rotorse.rp.AEP", "GW*h"],
        "LCOE": ["financese.lcoe", "USD/(MW*h)"],
        # "Cp": ["rotorse.rp.powercurve.Cp_aero", None],
        "Rated velocity": ["rotorse.rp.powercurve.rated_V", "m/s"],
        "Rated rpm": ["rotorse.rp.powercurve.rated_Omega", "rpm"],
        "Rated pitch": ["control.rated_pitch", "deg"],
        "Rated thrust": ["rotorse.rp.powercurve.rated_T", "kN"],
        "Rated torque": ["rotorse.rp.powercurve.rated_Q", "kN*m"],
        "Blade mass": ["rotorse.blade_mass", "kg"],
        "Blade cost": ["rotorse.re.precomp.total_blade_cost", "USD"],
        "Tip defl": ["tcons.tip_deflection", "m"],
        "Tip defl ratio": ["tcons.tip_deflection_ratio", None],
        "Flap freqs": ["rotorse.rs.frame.flap_mode_freqs", "Hz"],
        "Edge freqs": ["rotorse.rs.frame.edge_mode_freqs", "Hz"],
        "3P freq": ["rotorse.rp.powercurve.rated_Omega", None, 3.0 / 60],
        "6P freq": ["rotorse.rp.powercurve.rated_Omega", None, 6.0 / 60],
        "Hub forces": ["rotorse.rs.aero_hub_loads.Fxyz_hub_aero", "kN"],
        "Hub moments": ["rotorse.rs.aero_hub_loads.Mxyz_hub_aero", "kN*m"],
        "Nacelle mass": ["drivese.nacelle_mass", "kg"],
        "RNA mass": ["drivese.rna_mass", "kg"],
        "Tower mass": ["towerse.tower_mass", "kg"],
        "Floating Tower mass": ["floatingse.tower_mass", "kg"],
        "Tower cost": ["towerse.tower_cost", "USD"],
        "Floating Tower cost": ["floatingse.tower_cost", "USD"],
        "Monopile mass": ["towerse.monopile_mass", "kg"],
        "Monopile cost": ["towerse.monopile_cost", "USD"],
        "Tower freqs": ["towerse.tower.structural_frequencies", "Hz"],
        "Monopile/jacket freqs": ["fixedse.structural_frequencies", "Hz"],
        "Floating Tower freqs": ["floatingse.tower_freqs", "Hz"],
    }

    # Generally it's not necessary to change the code below here, unless you
    # want to plot additional values

    # Create the output folder if it doesn't exist
    if not os.path.exists(folder_output):
        os.makedirs(folder_output)

    # Call the functions to print, save, and plot results
    print_results_to_screen(list_of_sims, list_of_labels, values_to_print)
    save_lcoe_data_to_file(list_of_sims, folder_output)
    create_all_plots(
        list_of_sims,
        list_of_labels,
        modeling_options,
        analysis_options,
        folder_output,
        show_plots,
        font_size,
        extension,
    )


def main():
    # Called only if this script is run as main.

    # ======================================================================
    # Input Information
    # ======================================================================
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "input_files",
        nargs="*",
        type=str,
        help="Specify the yaml filenames and pickled output filenames to be compared.",
    )
    parser.add_argument(
        "--modeling_options",
        nargs="?",
        type=str,
        default=fname_modeling_options_default,
        help="Specify the modeling options yaml.",
    )
    parser.add_argument(
        "--analysis_options",
        nargs="?",
        type=str,
        default=fname_analysis_options_default,
        help="Specify the analysis options yaml.",
    )
    parser.add_argument(
        "--labels",
        nargs="*",
        type=str,
        default=None,
        help="Specify the labels for the yaml files (use spaces for separation, no brackets).",
    )

    args = parser.parse_args()
    input_filenames = args.input_files
    fname_modeling_options = args.modeling_options
    fname_analysis_options = args.analysis_options
    list_of_labels = args.labels

    if len(input_filenames) == 0:
        print("ERROR: Must specify either a set of yaml files or pkl files\n")
        parser.print_help(sys.stderr)
        sys.exit(1)

    if list_of_labels is None:
        list_of_labels = [f"run_{idx}" for idx in range(len(input_filenames))]

    list_of_sims = []
    for input_filename in input_filenames:
        froot = os.path.splitext(input_filename)[0]

        print()
        if os.path.exists(froot + ".yaml") and os.path.exists(froot + ".pkl"):
            # Load in saved pickle dictionary if already ran WISDEM
            print(f"Loading WISDEM data for {input_filename}.")
            wt_opt, modeling_options, analysis_options = load_wisdem(froot)

        else:
            # Run WISDEM for each yaml file to compare using the modeling and analysis options set above
            print(f"Running WISDEM for {input_filename}.")
            wt_opt, modeling_options, analysis_options = run_wisdem(
                input_filename,
                fname_modeling_options,
                fname_analysis_options,
                run_only=True,
            )

        list_of_sims.append(wt_opt)

    run(list_of_sims, list_of_labels, modeling_options, analysis_options)
    sys.exit(0)


if __name__ == "__main__":
    main()



import numpy as np

from wisdem.commonse.constants import eps
from wisdem.commonse.utilities import smooth_max, smooth_min, cubic_spline_eval
from scipy.optimize import fsolve


def hoopStressEurocode(d, t, L_reinforced, hoop):
    """default method for scaling hoop stress using Eurocode method
    GB 06/21/2018: Ansys comparisons for submerged case suggests this over-compensates for stiffener
                   I'm not even sure the Eurocode is implemented correctly here.  Suggest using the standard
                   hoop stress expression above or API's handling of ring stiffeners below.
    """

    r = d / 2.0 - t / 2.0  # radius of cylinder middle surface
    omega = L_reinforced / np.sqrt(r * t)

    C_theta = 1.5  # clamped-clamped
    k_w = 0.46 * (1.0 + 0.1 * np.sqrt(C_theta / omega * r / t))
    k_w, _, _ = smooth_max(k_w, 0.65)
    k_w, _, _ = smooth_min(k_w, 1.0)
    return k_w * hoop

def YieldCriterionEurocode(sigma_x, sigma_z, tau, gamma, sigma_y):
    """Yield criterion from EN1993 1-1 (6.1)
    """

    a = (sigma_x*gamma/sigma_y)**2
    b = (sigma_z*gamma/sigma_y)**2
    c = (sigma_x*gamma/sigma_y)*(sigma_z*gamma/sigma_y)
    d = 3*(tau*gamma/sigma_y)**2

    # stress margin   
    stress_utilization = a+b-c+d

    return stress_utilization


def bucklingGL(d, t, Fz, Myy, tower_height, E, sigma_y, gamma_f=1.2, gamma_b=1.1):
    # other factors
    alpha = 0.21  # buckling imperfection factor
    beta = 1.0  # bending coefficient
    sk_factor = 2.0  # fixed-free
    tower_height = tower_height * sk_factor

    # geometry
    A = np.pi * d * t
    I = np.pi * (d / 2.0) ** 3 * t
    Wp = I / (d / 2.0)

    # applied loads
    Nd = -Fz * gamma_f
    Md = Myy * gamma_f

    # plastic resistance
    Np = A * sigma_y / gamma_b
    Mp = Wp * sigma_y / gamma_b

    # factors
    Ne = np.pi**2 * (E * I) / (1.1 * tower_height**2)
    lambda_bar = np.sqrt(Np * gamma_b / Ne)
    phi = 0.5 * (1 + alpha * (lambda_bar - 0.2) + lambda_bar**2)
    kappa = np.ones(A.shape)
    idx = lambda_bar > 0.2
    kappa[idx] = 1.0 / (phi[idx] + np.sqrt(phi[idx] ** 2 - lambda_bar[idx] ** 2))
    delta_n = 0.25 * kappa * lambda_bar**2
    delta_n = np.minimum(delta_n, 0.1)

    GL_utilization = Nd / (kappa * Np) + beta * Md / Mp + delta_n  # this is utilization must be <1

    return GL_utilization


def shellBucklingEurocode(d, t, sigma_z, sigma_t, tau_zt, L_reinforced, E, sigma_y, gamma_f=1.2, gamma_b=1.1):
    """
    Estimate shell buckling utilization along tower.

    Arguments:
    npt - number of locations at each node at which stress is evaluated.
    sigma_z - axial stress at npt*node locations.  must be in order
                  [(node1_pts1-npt), (node2_pts1-npt), ...]
    sigma_t - azimuthal stress given at npt*node locations
    tau_zt - shear stress (z, theta) at npt*node locations
    E - modulus of elasticity
    sigma_y - yield stress
    L_reinforced - reinforcement length - structure is re-discretized with this spacing
    gamma_f - safety factor for stresses
    gamma_b - safety factor for buckling

    Returns:
    z
    EU_utilization: - array of shell buckling utilizations evaluted at (z[0] at npt locations, \n
                      z[0]+L_reinforced at npt locations, ...). \n
                      Each utilization must be < 1 to avoid failure.
    """

    n = len(t)
    EU_utilization = np.zeros(n)
    sigma_z_sh = np.zeros(n)
    sigma_t_sh = np.zeros(n)
    tau_zt_sh = np.zeros(n)

    for i in range(n):
        h = L_reinforced[i]

        r1 = d[i] / 2.0 - t[i] / 2.0
        r2 = d[i + 1] / 2.0 - t[i] / 2.0

        sigma_z_shell = sigma_z[i]
        sigma_t_shell = sigma_t[i]
        tau_zt_shell = tau_zt[i]

        # TODO: the following is non-smooth, although in general its probably OK
        # change to magnitudes and add safety factor
        sigma_z_shell = gamma_f * abs(sigma_z_shell)
        sigma_t_shell = gamma_f * abs(sigma_t_shell)
        tau_zt_shell = gamma_f * abs(tau_zt_shell)

        EU_utilization[i] = _shellBucklingOneSection(
            h, r1, r2, t[i], gamma_b, sigma_z_shell, sigma_t_shell, tau_zt_shell, E[i], sigma_y[i]
        )

        # make them into vectors
        sigma_z_sh[i] = sigma_z_shell
        sigma_t_sh[i] = sigma_t_shell
        tau_zt_sh[i] = tau_zt_shell

    return EU_utilization  # this is utilization must be <1


def _cxsmooth(omega, rovert):
    Cxb = 6.0  # clamped-clamped
    constant = 1 + 1.83 / 1.7 - 2.07 / 1.7**2

    ptL1 = 1.7 - 0.25
    ptR1 = 1.7 + 0.25

    ptL2 = 0.5 * rovert - 1.0
    ptR2 = 0.5 * rovert + 1.0

    ptL3 = (0.5 + Cxb) * rovert - 1.0
    ptR3 = (0.5 + Cxb) * rovert + 1.0

    if omega < ptL1:
        Cx = constant - 1.83 / omega + 2.07 / omega**2

    elif omega >= ptL1 and omega <= ptR1:
        fL = constant - 1.83 / ptL1 + 2.07 / ptL1**2
        fR = 1.0
        gL = 1.83 / ptL1**2 - 4.14 / ptL1**3
        gR = 0.0
        Cx = cubic_spline_eval(ptL1, ptR1, fL, fR, gL, gR, omega)

    elif omega > ptR1 and omega < ptL2:
        Cx = 1.0

    elif omega >= ptL2 and omega <= ptR2:
        fL = 1.0
        fR = 1 + 0.2 / Cxb * (1 - 2.0 * ptR2 / rovert)
        gL = 0.0
        gR = -0.4 / Cxb / rovert
        Cx = cubic_spline_eval(ptL2, ptR2, fL, fR, gL, gR, omega)

    elif omega > ptR2 and omega < ptL3:
        Cx = 1 + 0.2 / Cxb * (1 - 2.0 * omega / rovert)

    elif omega >= ptL3 and omega <= ptR3:
        fL = 1 + 0.2 / Cxb * (1 - 2.0 * ptL3 / rovert)
        fR = 0.6
        gL = -0.4 / Cxb / rovert
        gR = 0.0
        Cx = cubic_spline_eval(ptL3, ptR3, fL, fR, gL, gR, omega)

    else:
        Cx = 0.6

    return Cx


def _sigmasmooth(omega, E, rovert):
    Ctheta = 1.5  # clamped-clamped

    ptL = 1.63 * rovert * Ctheta - 1
    ptR = 1.63 * rovert * Ctheta + 1

    if omega < 20.0 * Ctheta:
        offset = 10.0 / (20 * Ctheta) ** 2 - 5 / (20 * Ctheta) ** 3
        Cthetas = 1.5 + 10.0 / omega**2 - 5 / omega**3 - offset
        sigma = 0.92 * E * Cthetas / omega / rovert

    elif omega >= 20.0 * Ctheta and omega < ptL:
        sigma = 0.92 * E * Ctheta / omega / rovert

    elif omega >= ptL and omega <= ptR:
        alpha1 = 0.92 / 1.63 - 2.03 / 1.63**4

        fL = 0.92 * E * Ctheta / ptL / rovert
        fR = E * (1.0 / rovert) ** 2 * (alpha1 + 2.03 * (Ctheta / ptR * rovert) ** 4)
        gL = -0.92 * E * Ctheta / rovert / ptL**2
        gR = -E * (1.0 / rovert) * 2.03 * 4 * (Ctheta / ptR * rovert) ** 3 * Ctheta / ptR**2

        sigma = cubic_spline_eval(ptL, ptR, fL, fR, gL, gR, omega)

    else:
        alpha1 = 0.92 / 1.63 - 2.03 / 1.63**4
        sigma = E * (1.0 / rovert) ** 2 * (alpha1 + 2.03 * (Ctheta / omega * rovert) ** 4)

    return sigma


def _tausmooth(omega, rovert):
    ptL1 = 9
    ptR1 = 11

    ptL2 = 8.7 * rovert - 1
    ptR2 = 8.7 * rovert + 1

    if omega < ptL1:
        C_tau = np.sqrt(1.0 + 42.0 / omega**3 - 42.0 / 10**3)

    elif omega >= ptL1 and omega <= ptR1:
        fL = np.sqrt(1.0 + 42.0 / ptL1**3 - 42.0 / 10**3)
        fR = 1.0
        gL = -63.0 / ptL1**4 / fL
        gR = 0.0
        C_tau = cubic_spline_eval(ptL1, ptR1, fL, fR, gL, gR, omega)

    elif omega > ptR1 and omega < ptL2:
        C_tau = 1.0

    elif omega >= ptL2 and omega <= ptR2:
        fL = 1.0
        fR = 1.0 / 3.0 * np.sqrt(ptR2 / rovert) + 1 - np.sqrt(8.7) / 3
        gL = 0.0
        gR = 1.0 / 6 / np.sqrt(ptR2 * rovert)
        C_tau = cubic_spline_eval(ptL2, ptR2, fL, fR, gL, gR, omega)

    else:
        C_tau = 1.0 / 3.0 * np.sqrt(omega / rovert) + 1 - np.sqrt(8.7) / 3

    return C_tau


def _shellBucklingOneSection(h, r1, r2, t, gamma_b, sigma_z, sigma_t, tau_zt, E, sigma_y):
    """
    Estimate shell buckling for one tapered cylindrical shell section.

    Arguments:
    h - height of conical section
    r1 - radius at bottom
    r2 - radius at top
    t - shell thickness
    E - modulus of elasticity
    sigma_y - yield stress
    gamma_b - buckling reduction safety factor
    sigma_z - axial stress component
    sigma_t - azimuthal stress component
    tau_zt - shear stress component (z, theta)

    Returns:
    EU_utilization, shell buckling utilization which must be < 1 to avoid failure

    """

    # NOTE: definition of r1, r2 switched from Eurocode document to be consistent with FEM.

    # ----- geometric parameters --------
    beta = np.arctan2(r1 - r2, h)
    L = h / np.cos(beta)

    # ------------- axial stress -------------
    # length parameter
    le = L
    re = 0.5 * (r1 + r2) / np.cos(beta)
    omega = le / np.sqrt(re * t)
    rovert = re / t

    # compute Cx
    Cx = _cxsmooth(omega, rovert)

    # if omega <= 1.7:
    #     Cx = 1.36 - 1.83/omega + 2.07/omega/omega
    # elif omega > 0.5*rovert:
    #     Cxb = 6.0  # clamped-clamped
    #     Cx = max(0.6, 1 + 0.2/Cxb*(1-2.0*omega/rovert))
    # else:
    #     Cx = 1.0

    # critical axial buckling stress
    sigma_z_Rcr = 0.605 * E * Cx / rovert

    # compute buckling reduction factors
    lambda_z0 = 0.2
    beta_z = 0.6
    eta_z = 1.0
    Q = 25.0  # quality parameter - high
    lambda_z = np.sqrt(sigma_y / sigma_z_Rcr)
    delta_wk = 1.0 / Q * np.sqrt(rovert) * t
    alpha_z = 0.62 / (1 + 1.91 * (delta_wk / t) ** 1.44)

    chi_z = _buckling_reduction_factor(alpha_z, beta_z, eta_z, lambda_z0, lambda_z)

    # design buckling stress
    sigma_z_Rk = chi_z * sigma_y
    sigma_z_Rd = sigma_z_Rk / gamma_b

    # ---------------- hoop stress ------------------

    # length parameter
    le = L
    re = 0.5 * (r1 + r2) / (np.cos(beta))
    omega = le / np.sqrt(re * t)
    rovert = re / t

    # Ctheta = 1.5  # clamped-clamped
    # CthetaS = 1.5 + 10.0/omega**2 - 5.0/omega**3

    # # critical hoop buckling stress
    # if (omega/Ctheta < 20.0):
    #     sigma_t_Rcr = 0.92*E*CthetaS/omega/rovert
    # elif (omega/Ctheta > 1.63*rovert):
    #     sigma_t_Rcr = E*(1.0/rovert)**2*(0.275 + 2.03*(Ctheta/omega*rovert)**4)
    # else:
    #     sigma_t_Rcr = 0.92*E*Ctheta/omega/rovert

    sigma_t_Rcr = np.maximum(eps, _sigmasmooth(omega, E, rovert))

    # buckling reduction factor
    alpha_t = 0.65  # high fabrication quality
    lambda_t0 = 0.4
    beta_t = 0.6
    eta_t = 1.0
    lambda_t = np.sqrt(sigma_y / sigma_t_Rcr)

    chi_theta = _buckling_reduction_factor(alpha_t, beta_t, eta_t, lambda_t0, lambda_t)

    sigma_t_Rk = chi_theta * sigma_y
    sigma_t_Rd = sigma_t_Rk / gamma_b

    # ----------------- shear stress ----------------------

    # length parameter
    le = h
    rho = np.sqrt((r1 + r2) / (2.0 * r2))
    re = (1.0 + rho - 1.0 / rho) * r2 * np.cos(beta)
    omega = le / np.sqrt(re * t)
    rovert = re / t

    # if (omega < 10):
    #     C_tau = np.sqrt(1.0 + 42.0/omega**3)
    # elif (omega > 8.7*rovert):
    #     C_tau = 1.0/3.0*np.sqrt(omega/rovert)
    # else:
    #     C_tau = 1.0
    C_tau = _tausmooth(omega, rovert)

    tau_zt_Rcr = 0.75 * E * C_tau * np.sqrt(1.0 / omega) / rovert

    # reduction factor
    alpha_tau = 0.65  # high fabrifaction quality
    beta_tau = 0.6
    lambda_tau0 = 0.4
    eta_tau = 1.0
    lambda_tau = np.sqrt(sigma_y / np.sqrt(3) / tau_zt_Rcr)

    chi_tau = _buckling_reduction_factor(alpha_tau, beta_tau, eta_tau, lambda_tau0, lambda_tau)

    tau_zt_Rk = chi_tau * sigma_y / np.sqrt(3)
    tau_zt_Rd = tau_zt_Rk / gamma_b

    # buckling interaction parameters

    k_z = 1.25 + 0.75 * chi_z
    k_theta = 1.25 + 0.75 * chi_theta
    k_tau = 1.75 + 0.25 * chi_tau
    k_i = (chi_z * chi_theta) ** 2

    # shell buckling utilization

    utilization = (
        (sigma_z / sigma_z_Rd) ** k_z
        + (sigma_t / sigma_t_Rd) ** k_theta
        - k_i * (sigma_z * sigma_t / sigma_z_Rd / sigma_t_Rd)
        + (tau_zt / tau_zt_Rd) ** k_tau
    )

    return utilization  # this is utilization must be <1


def _buckling_reduction_factor(alpha, beta, eta, lambda_0, lambda_bar):
    """
    Computes a buckling reduction factor used in Eurocode shell buckling formula.
    """

    lambda_p = np.sqrt(alpha / (1.0 - beta))

    ptL = 0.9 * lambda_0
    ptR = 1.1 * lambda_0

    if lambda_bar < ptL:
        chi = 1.0

    elif lambda_bar >= ptL and lambda_bar <= ptR:  # cubic spline section
        fracR = (ptR - lambda_0) / (lambda_p - lambda_0)
        fL = 1.0
        fR = 1 - beta * fracR**eta
        gL = 0.0
        gR = -beta * eta * fracR ** (eta - 1) / (lambda_p - lambda_0)

        chi = cubic_spline_eval(ptL, ptR, fL, fR, gL, gR, lambda_bar)

    elif lambda_bar > ptR and lambda_bar < lambda_p:
        chi = 1.0 - beta * ((lambda_bar - lambda_0) / (lambda_p - lambda_0)) ** eta

    else:
        chi = alpha / lambda_bar**2

    # if (lambda_bar <= lambda_0):
    #     chi = 1.0
    # elif (lambda_bar >= lambda_p):
    #     chi = alpha/lambda_bar**2
    # else:
    #     chi = 1.0 - beta*((lambda_bar-lambda_0)/(lambda_p-lambda_0))**eta

    return chi


class memberBuckling:
    """
    class for calculating member buckling based on EN 1993 1-1

    """

    def __init__(
        self,
        a,
        b,
        l,
        A,
        Ixx,
        Iyy,
        Fz,
        Mx,
        My,
        I_T,
        E=200e9,
        G=79.3e9,
        sigma_y=345e6,
        gamma_m=1.1,
        kw=0.5,
        Iw=0,
        k=0.5,
        **kwargs,
    ):
        """
        create an instance of memberBuckling

        Args:
            l   (np.array): section length
            A   (np.array): Areas of rectangular member section
            Ixx (np.array): Moment of inertia x
            Iyy (np.array): Moment of inertia y
            IT (float): St Venent torsional constant
            kw (float): factor allowing for end constraint, default to full fixity 0.5
            E   (float): Isotropic Young's modulus [Pa]. Defaults to 200e9.
            G   (float, optional): Shear modulus [Pa]. Defaults to 79.3e9.
            sigma_y (float): Yield strength [Pa]. Defaults to 345e6.
            gamma_m (float, optional): Partial factor for resistance of members to instability. Defaults to 1.1 by DNVGL.
        """
        self.a = a
        self.b = b
        self.A = A
        self.Ixx = Ixx
        self.Iyy = Iyy
        self.Fz = Fz
        self.Mx = Mx
        self.My = My
        self.E = E
        self.l = l
        self.I_T = I_T
        self.Iw = Iw
        self.k = k
        self.kw = kw
        self.G = G
        self.sigma_y = sigma_y
        self.gamma_m = gamma_m

        self.k = 0.5 # factor due to end constraint
        # The design mormal force is compression
        self.Ned = Fz # Might need to do Fz[:-1]-Fz[1:] as axial loads

        # Elastic section modulus =  ratio of the second moment of area and the distance from the neutral axis
        # to any given fiber
        self.Welx = self.Ixx/(self.a/2)
        self.Wely = self.Iyy/(self.b/2)

        # gyration radius
        self.ix = np.sqrt((self.Ixx/self.A))
        self.iy = np.sqrt((self.Iyy/self.A))

        # distance to shear center
        self.x0 = 0
        self.y0 = 0

        self.elastic_section_modulus()
        self.critical_loads()
        self.characteristic_resistance()
        self.reduction_factor()
        self.interaction_factor()
        

    def critical_loads(self):

        self.Ncrx = np.pi**2*self.E*self.Ixx/(self.k*self.l)
        self.Ncry = np.pi**2*self.E*self.Iyy/(self.k*self.l)
        # NcrT: elastic torsional-flexural buckling force
        # Reference: https://help.scia.net/19.1/en/pvt/steelcodechecktb/en/torsional_flexural_buckling.htm
        self.i02 = self.ix**2+self.iy**2+self.x0**2+self.y0**2
        self.NcrT = 1.0/self.i02*(self.G*self.I_T+np.pi**2*self.E*self.Iw/(self.k*self.l)**2)
        fak = fsolve(self.NcrTF_eqn, self.NcrT, xtol=1e-4, maxfev=50)
        self.NcrTF = fak

        # I_T is the St Venent torsinal constant
        # For rectangular thin-walled with uniform thickness I_T = 2*t**2*(a-t)**2*(b-t)**2/(a*t+b*t-2*t**2)
        # Iw is the warping constant
        # For rectangular, Iw is usually zero
        # kw is factor allowing for end constraint, refers to end support warping, 1 for no fixity and 0.5 for full fixity
        self.Mcrx = self.Ncrx*np.sqrt(self.k/self.kw*self.Iw/self.Ixx+1/self.Ncrx*self.G*self.I_T/self.Ixx)
        self.Mcry = self.Ncry*np.sqrt(self.k/self.kw*self.Iw/self.Iyy+1/self.Ncry*self.G*self.I_T/self.Iyy)

    def NcrTF_eqn(self, Nload):
        return self.i02*(Nload-self.Ncrx)*(Nload-self.NcrT) - Nload**2*self.y0**2*(Nload-self.Ncrx) - Nload**2*self.x0**2*(Nload-self.Ncry)

    def elastic_section_modulus(self):
        self.Wx = self.Ixx*2/self.b
        self.Wy = self.Iyy*2/self.a

    def characteristic_resistance(self):
        self.Nrk = self.A*self.sigma_y
        self.Mxrk = self.Ixx*self.sigma_y
        self.Myrk = self.Iyy*self.sigma_y

    def interaction_factor(self):
        alpha_LT_x = np.max(1-self.I_T/self.Ixx, 0)
        alpha_LT_y = np.max(1-self.I_T/self.Iyy, 0)

        # Equivalent uniform moment, EN 1993 1-1 Table A.2
        # Psi_x = self.Mx[1:]/self.Mx[:-1]
        # Psi_y = self.My[1:]/self.My[:-1]
        Psi_x = np.ones(len(self.a))
        Psi_y = np.ones_like(Psi_x)

        cmx0 = 0.79+0.21*Psi_x+0.36*(Psi_x-0.33)*self.Ned/self.Ncrx
        cmy0 = 0.79+0.21*Psi_y+0.36*(Psi_y-0.33)*self.Ned/self.Ncry

        mux = (1-self.Ned/self.Ncrx)/(1-self.Chi_x*self.Ned/self.Ncrx)
        muy = (1-self.Ned/self.Ncry)/(1-self.Chi_y*self.Ned/self.Ncry)


        kc_x = 1/(1.33-0.33*Psi_x)
        kc_y = 1/(1.33-0.33*Psi_y)

        C1_x = kc_x**(-2) # is a factor depending on the loading and end conditions
        C1_y = kc_y**(-2) # is a factor depending on the loading and end conditions

        lambda_crit_x = 0.2*np.sqrt(C1_x)*((1-self.Ned/self.Ncry)*(1-self.Ned/self.NcrT))**0.25
        lambda_crit_y = 0.2*np.sqrt(C1_x)*((1-self.Ned/self.Ncrx)*(1-self.Ned/self.NcrT))**0.25

        lambda_0_x =  0.2*((1-self.Ned/self.Ncry)*(1-self.Ned/self.NcrT))**0.25
        lambda_0_y =  0.2*((1-self.Ned/self.Ncrx)*(1-self.Ned/self.NcrT))**0.25

        # Take the moments directly as the maximum design values of bending moments
        Mxed = self.Mx
        Myed = self.My

        esp_x = Mxed/self.Ned*self.A/self.Welx
        esp_y = Myed/self.Ned*self.A/self.Wely

        self.kxx = np.zeros_like(self.Ned)
        self.kyy = np.zeros_like(self.kxx)
        self.kxy = np.zeros_like(self.kxx)
        self.kyx = np.zeros_like(self.kxx)


        for i in range(len(cmx0)):

            if lambda_0_x[i] <= lambda_crit_x[i]:
                cmx = cmx0[i]
                cmy = cmy0[i]
                cmLT=1.0

            else:
                cmx = cmx0[i]+(1-cmx0[i])*np.sqrt(esp_x[i])*alpha_LT_x[i]/(1+np.sqrt(esp_x[i])*alpha_LT_x[i])
                cmy = cmy0[i]
                cmLT = cmy**2*alpha_LT_x[i]/(np.sqrt((1-self.Ned[i]/self.Ncry[i])*(1-self.Ned[i]/self.NcrTF[i])))

            self.kxx[i] = cmx*cmLT*mux[i]/(1-self.Ned[i]/self.Ncrx[i])
            self.kxy[i] = cmy*mux[i]/(1-self.Ned[i]/self.Ncry[i])

            if lambda_0_y[i] <= lambda_crit_y[i]:
                cmx = cmx0[i]
                cmy = cmy0[i]
                cmLT=1.0

            else:
                cmx = cmx0[i]
                cmy = cmy0[i]+(1-cmy0[i])*np.sqrt(esp_y[i])*alpha_LT_y[i]/(1+np.sqrt(esp_y[i])*alpha_LT_y[i])
                cmLT = cmy**2*alpha_LT_y[i]/(np.sqrt((1-self.Ned[i]/self.Ncrx[i])*(1-self.Ned[i]/self.NcrTF[i])))

            self.kyx[i] = cmx*cmLT*muy[i]/(1-self.Ned[i]/self.Ncrx[i])
            self.kyy[i] = cmy*muy[i]/(1-self.Ned[i]/self.Ncry[i])

    def _reduction_factor(self, lambda_slender, Psi, alpha_imperf=0.49):
        """
        Compute the reduction factor due to flexural buckling or lateral torsion buckling

        Args:
            lambda_slender (float): Slenderness due to flexural buckling or lateral torsion
            alpha_imperf (float): Imperfection factor, default 0.49, EN1993 1-1 Table 6.1 and 6.2
        retuun:
            reduction factors
        """

        Phi = 0.5*(1+alpha_imperf*(lambda_slender-0.2)+lambda_slender**2)
        Chi = 1/(Phi+np.sqrt(Phi**2-lambda_slender**2))

        # kc is the correction factor
        # Taking into account the moment distribution between the lateral restraints of members
        # See Table 6.6 in Eurocode 
        # the reduction factor may be modified as follows
        # Only consider linear variation
        kc = 1/(1.33-0.33*Psi)

        f = 1.0 - 0.5*(1-kc)*(1-2.0*(lambda_slender-0.8)**2)

        return Chi
    
    def reduction_factor(self, alpha_imperf=0.49):

        # Reduction factors
        # Because the moments are on the sections, there is not end moment compute Psi
        # Currecntly assume uniform distribution, so Psi=1
        # Psi_x = self.Mx[1:]/self.Mx[:-1]
        # Psi_y = self.My[1:]/self.My[:-1]
        Psi_x = np.ones(len(self.a))
        Psi_y = np.ones_like(Psi_x)
        lambda_slender_x, lambda_slender_y = self.flexural_slender()
        self.Chi_x = self._reduction_factor(lambda_slender_x, Psi=Psi_x, alpha_imperf=alpha_imperf)
        self.Chi_y = self._reduction_factor(lambda_slender_y, Psi=Psi_y, alpha_imperf=alpha_imperf)
        lambda_slender_x, lambda_slender_y = self.lateral_torsion_slender()
        self.Chi_LT_x = self._reduction_factor(lambda_slender_x, Psi=Psi_x, alpha_imperf=alpha_imperf)
        self.Chi_LT_y = self._reduction_factor(lambda_slender_y, Psi=Psi_y, alpha_imperf=alpha_imperf)


    def utilization(self, Ned, N_rk, Mx_ed, Mx_rk, My_ed, My_rk):
        """
        Compute the buckling constraints

        Args:
            Ned (np.array): Design values of the compression force and the maximum moments about the x-x and y-y axes
            N_rk (float): Characteristic resistance to normal force of the critical cross sections
            Mx_ed (float):  Design bending moment
            Mx_rk (float): Characteristic moment resistance of the critical cross sections
            My_ed (float): Design bending moment
            My_rk (float): Characteristic moment resistance of the critical cross sections
        """
        # Ignore the shift of the centroidal axes
        # The subscript of force and moment are modified from EN 1993 definition to be consistent within
        # WISDEM

        a = Ned*self.gamma_m/self.Chi_x/N_rk
        b = self.kxx*Mx_ed*self.gamma_m/self.Chi_LT_x/Mx_rk
        c = self.kxy*My_ed*self.gamma_m/self.Chi_LT_x/My_rk

        utilization_x = a + b + c

        a = Ned*self.gamma_m/self.Chi_y/N_rk
        b = self.kyx*Mx_ed*self.gamma_m/self.Chi_LT_y/Mx_rk
        c = self.kyy*My_ed*self.gamma_m/self.Chi_LT_y/My_rk

        utilization_y = a + b + c

        return utilization_x, utilization_y
    
    def run_buckling_checks(self, Fz, Mxx, Myy):

        Ned = Fz
        N_rk = self.A*self.sigma_y/self.gamma_m
        Mx_ed = Mxx
        My_ed = Myy
        Mx_rk = self.Welx*self.sigma_y/self.gamma_m
        My_rk = self.Wely*self.sigma_y/self.gamma_m

        utilization_x, utilization_y = self.utilization(Ned, N_rk, Mx_ed, Mx_rk, My_ed, My_rk)

        return np.maximum(utilization_x, utilization_y)

    
    def flexural_slender(self):

        lambda_slender_x = np.sqrt(self.A*self.sigma_y/self.Ncrx)
        lambda_slender_y = np.sqrt(self.A*self.sigma_y/self.Ncry)
        
        return lambda_slender_x, lambda_slender_y
    
    def lateral_torsion_slender(self):

        lambda_slender_x = np.sqrt(self.Wx*self.sigma_y/self.Mcrx)
        lambda_slender_y = np.sqrt(self.Wy*self.sigma_y/self.Mcry)

        return lambda_slender_x, lambda_slender_y
    

